# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import mock
import six

from neutron.tests import base

from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.common import constants
from networking_cisco.apps.saf.db import dfa_db_models as dbm
from networking_cisco.apps.saf.server import cisco_dfa_rest as cdr
from networking_cisco.apps.saf.server import dfa_events_handler as deh
from networking_cisco.apps.saf.server import dfa_fail_recovery as dfr
from networking_cisco.apps.saf.server import dfa_instance_api as dia
from networking_cisco.apps.saf.server.services.firewall.native import (
    fw_mgr as fw_native)

FAKE_NETWORK_NAME = 'test_dfa_network'
FAKE_NETWORK_ID = '949fdd05-a26a-4819-a829-9fc2285de6ff'
FAKE_NETWORK_ID2 = '949fdd05-a26a-4819-a829-9fc2285de6fe'
FAKE_CFG_PROF_ID = '8c30f360ffe948109c28ab56f69a82e1'
FAKE_SEG_ID = 12345
FAKE_PROJECT_NAME = 'test_dfa_project'
FAKE_ORCH_ID = 'openstack'
FAKE_PROJECT_ID = 'aee5da7e699444889c662cf7ec1c8de7'
FAKE_PROJECT_ID2 = 'aee5da7e699444889c662cf7ec1c8d8'
FAKE_PROJECT_ID3 = 'aee5da7e699444889c662cf7ec1c8d9'
FAKE_CFG_PROFILE_NAME = 'defaultNetworkL2Profile'
FAKE_INSTANCE_NAME = 'test_dfa_instance'
FAKE_SUBNET_ID = '1a3c5ee1-cb92-4fd8-bff1-8312ac295d64'
FAKE_PORT_ID = 'ea0d92cf-d0cb-4ed2-bbcf-ed7c6aaea4cb'
FAKE_DEVICE_ID = '20305657-78b7-48f4-a7cd-1edf3edbfcad'
FAKE_SECURITY_GRP_ID = '4b5b387d-cf21-4594-b926-f5a5c602295f'
FAKE_MAC_ADDR = 'fa:16:3e:70:15:c4'
FAKE_IP_ADDR = '23.24.25.4'
FAKE_GW_ADDR = '23.24.25.1'
FAKE_DHCP_IP_START = '23.24.25.2'
FAKE_DHCP_IP_END = '23.24.25.254'
FAKE_HOST_ID = 'test_dfa_host'
FAKE_FWD_MODE = 'proxy-gateway'
FAKE_DCNM_USERNAME = 'cisco'
FAKE_DCNM_PASSWD = 'password'
FAKE_DCNM_IP = '1.1.2.2'


class FakeClass(object):
    """Fake class"""
    @classmethod
    def imitate(cls, *others):
        for other in others:
            for name in other.__dict__:
                try:
                    setattr(cls, name, mock.Mock())
                except (TypeError, AttributeError):
                    pass
        return cls


class FakeProject(object):
    """Fake Project class."""
    def __init__(self, proj_id, name, dci_id, desc):
        self.id = proj_id
        self.name = name
        self.dci_id = dci_id
        self.description = desc


class TestDFAServer(base.BaseTestCase):
    """Test cases for DFA Server class."""

    def setUp(self):
        super(TestDFAServer, self).setUp()

        # Mocking some modules
        self.dcnmpatcher = mock.patch(cdr.__name__ + '.DFARESTClient')
        self.mdcnm = self.dcnmpatcher.start()

        self.keys_patcher = mock.patch(deh.__name__ + '.EventsHandler')
        self.mkeys = self.keys_patcher.start()

        self.inst_api_patcher = mock.patch(dia.__name__ + '.DFAInstanceAPI')
        self.m_inst_api = self.inst_api_patcher.start()

        self.module_patcher = mock.patch.dict('sys.modules',
                                              {'pika': mock.Mock()})
        self.module_patcher.start()

        from networking_cisco.apps.saf.server import dfa_listen_dcnm as dld
        from networking_cisco.apps.saf.server import dfa_server as ds
        self.dld_patcher = mock.patch(dld.__name__ + '.DCNMListener')
        self.dld = self.dld_patcher.start()

        ds.DfaServer.__bases__ = (FakeClass.imitate(
            dfr.DfaFailureRecovery, dbm.DfaDBMixin, fw_native.FwMgr),)

        ds.DfaServer.get_all_projects.return_value = []
        ds.DfaServer.get_all_networks.return_value = []
        ds.DfaServer._setup_rpc = mock.Mock()
        # TODO(padkrish) Have UT for this function. This may mean over-riding
        # the mocking of get_segmentid_range of DCNM client to return a range.
        ds.DfaServer.register_segment_dcnm = mock.Mock()
        # Setting DCNM parameters.
        config.default_dcnm_opts['dcnm']['dcnm_ip'] = FAKE_DCNM_IP
        config.default_dcnm_opts['dcnm']['dcnm_user'] = FAKE_DCNM_USERNAME
        config.default_dcnm_opts['dcnm']['dcnm_password'] = FAKE_DCNM_PASSWD
        config.default_dcnm_opts['dcnm']['timeout_resp'] = 0.01
        config.default_dcnm_opts['dcnm']['segmentation_id_min'] = 10000
        config.default_dcnm_opts['dcnm']['segmentation_id_max'] = 20000
        config.default_dcnm_opts['dcnm']['orchestrator_id'] = FAKE_ORCH_ID
        self.cfg = config.CiscoDFAConfig().cfg
        self.segid = int(self.cfg.dcnm.segmentation_id_min) + 10
        self.seg_Drvr = mock.patch(
            'networking_cisco.apps.saf.db.dfa_db_models.'
            'DfaSegmentTypeDriver').start()
        self.topologyDb = mock.patch(
            'networking_cisco.apps.saf.db.dfa_db_models.'
            'TopologyDiscoveryDb').start()

        self.dfa_server = ds.DfaServer(self.cfg)
        self.rpcb = ds.RpcCallBacks(self.dfa_server)
        mock.patch.object(self.dfa_server, '_get_segmentation_id',
                          return_value=12345).start()
        mock.patch.object(self.dfa_server.seg_drvr,
                          'allocate_segmentation_id',
                          return_value=12345).start()
        self.dciid = str(123)
        self.proj_desc = 'Unit Test Project'
        projs = [
            FakeProject(FAKE_PROJECT_ID, FAKE_PROJECT_NAME,
                        self.dciid, self.proj_desc)]
        self.dfa_server.get_all_projects.return_value = projs
        self.dfa_server._load_project_info_cache()
        self.part_name = self.cfg.dcnm.default_partition_name

    def _get_port_info(self):
        port_info = {'port': {
            'status': 'ACTIVE',
            'binding:host_id': FAKE_HOST_ID,
            'allowed_address_pairs': [],
            'extra_dhcp_opts': [],
            'device_owner': 'compute:nova',
            'binding:profile': {},
            'fixed_ips': [{'subnet_id': FAKE_SUBNET_ID,
                           'ip_address': FAKE_IP_ADDR}],
            'id': FAKE_PORT_ID,
            'security_groups': [FAKE_SECURITY_GRP_ID],
            'device_id': FAKE_DEVICE_ID,
            'name': '',
            'admin_state_up': True,
            'network_id': FAKE_NETWORK_ID,
            'tenant_id': FAKE_PROJECT_ID,
            'binding:vif_details': {'port_filter': True,
                                    'ovs_hybrid_plug': True},
            'binding:vnic_type': 'normal',
            'binding:vif_type': 'ovs',
            'mac_address': FAKE_MAC_ADDR}}
        return port_info

    def _load_network_info(self):
        dnet = mock.Mock()
        dnet.network_id = FAKE_NETWORK_ID
        dnet.segmentation_id = self.segid
        dnet.config_profile = FAKE_CFG_PROFILE_NAME
        dnet.fwd_mod = FAKE_FWD_MODE
        dnet.tenant_id = FAKE_PROJECT_ID
        dnet.name = FAKE_NETWORK_NAME
        self.dfa_server.get_all_networks.return_value = [dnet]
        self.dfa_server._load_network_info()

    def test_update_project_info_cache(self):
        """Test case for update project info."""

        pid = FAKE_PROJECT_ID
        name = FAKE_PROJECT_NAME
        dciid = 1000
        result = constants.RESULT_SUCCESS
        self.dfa_server.update_project_info_cache(pid, dci_id=dciid,
                                                  name=name, opcode='add')
        self.assertTrue(self.dfa_server.add_project_db.called)
        self.assertFalse(self.dfa_server.update_project_entry.called)
        self.assertFalse(self.dfa_server.del_project_db.called)
        self.dfa_server.add_project_db.assert_called_with(pid, name,
                                                          dciid, result)

        self.dfa_server.update_project_info_cache(pid, dci_id=dciid,
                                                  name=name,
                                                  opcode='update')
        self.assertTrue(self.dfa_server.update_project_entry.called)
        self.assertFalse(self.dfa_server.del_project_db.called)
        self.dfa_server.update_project_entry.assert_called_with(pid, dciid,
                                                                result)

    def test_project_create_func(self):
        """Test case for project create event."""

        dciid = str(12345)
        proj_desc = 'Unit Test Project'
        proj_id = FAKE_PROJECT_ID2
        proj = mock.Mock()
        proj.name = FAKE_PROJECT_NAME
        proj.description = proj_desc
        part_name = self.cfg.dcnm.default_partition_name
        self.dfa_server.keystone_event._service.projects.get.return_value = (
            proj)

        self.dfa_server.project_create_func(proj_id)
        # Try it with DCI id
        proj.name = FAKE_PROJECT_NAME + ':dci_id:' + dciid
        proj_id = FAKE_PROJECT_ID3
        self.dfa_server.project_create_func(proj_id)
        expected_calls = [
            mock.call(FAKE_ORCH_ID, FAKE_PROJECT_NAME, part_name, None,
                      proj.description),
            mock.call(FAKE_ORCH_ID, FAKE_PROJECT_NAME, part_name, dciid,
                      proj.description)]
        self.assertEqual(
            expected_calls,
            self.dfa_server.dcnm_client.create_project.call_args_list)

    def test_project_update_event(self):
        """Test case for project update event."""

        proj_info = {'resource_info': FAKE_PROJECT_ID}
        proj = mock.Mock()
        proj.name = FAKE_PROJECT_NAME + ':dci_id:' + self.dciid
        proj.description = self.proj_desc
        self.dfa_server.keystone_event._service.projects.get.return_value = (
            proj)

        self.dfa_server.project_update_event(proj_info)
        # Project update event is called with the same parameters. It is
        # expected that there is no call to update_project.
        self.assertFalse(
            self.dfa_server.dcnm_client.update_project.called)

        # Try with updating the project by name.
        proj.name = FAKE_PROJECT_NAME + 'new' + ':dci_id:' + self.dciid
        self.dfa_server.project_update_event(proj_info)
        self.assertFalse(
            self.dfa_server.dcnm_client.update_project.called)

        # Try with updating the dci_id of the project.
        proj.name = FAKE_PROJECT_NAME + ':dci_id:' + str(124)
        self.dfa_server.project_update_event(proj_info)
        self.assertTrue(
            self.dfa_server.dcnm_client.update_project.called)
        expected_calls = [mock.call(FAKE_PROJECT_NAME,
                                    self.cfg.dcnm.default_partition_name,
                                    dci_id=str(124))]
        self.assertEqual(
            expected_calls,
            self.dfa_server.dcnm_client.update_project.call_args_list)

    def test_project_delete_event(self):
        """Test case for project delete event."""

        proj_name = FAKE_PROJECT_NAME
        proj_info = {'resource_info': FAKE_PROJECT_ID}
        part_name = self.cfg.dcnm.default_partition_name

        self.dfa_server.project_delete_event(proj_info)

        # Check information sent to dcnm and api that deleting the entry from
        # DB is called.
        self.dfa_server.dcnm_client.delete_project.assert_called_with(
            proj_name, part_name)

        self.dfa_server.del_project_db.assert_called_with(FAKE_PROJECT_ID)

    def test_network_create_func(self):
        """Test case for network create event."""

        net = {'name': FAKE_NETWORK_NAME,
               'tenant_id': FAKE_PROJECT_ID,
               'id': FAKE_NETWORK_ID2}
        dcnmclnt = self.dfa_server.dcnm_client
        dcnmclnt.get_config_profile_for_network.return_value = (
            FAKE_CFG_PROFILE_NAME, FAKE_FWD_MODE)
        self.dfa_server.get_network.return_value = None
        self.dfa_server.network_create_func(net)
        dfa_net = self.dfa_server.network[FAKE_NETWORK_ID2]
        expected_calls = [mock.call(FAKE_NETWORK_ID2, dfa_net, 'openstack',
                                    constants.SUBNET_PENDING)]
        self.assertEqual(expected_calls,
                         self.dfa_server.add_network_db.call_args_list)

    def test_subnet_create_event(self):
        """Test case for subnet create event."""

        network_info = {'network':
                        {'name': FAKE_NETWORK_NAME,
                         'tenant_id': FAKE_PROJECT_ID,
                         'id': FAKE_NETWORK_ID2}}
        subnet_info = {'subnet': {
            'network_id': FAKE_NETWORK_ID2,
            'tenant_id': FAKE_PROJECT_ID,
            'allocation_pools': [
                {'start': FAKE_DHCP_IP_START, 'end': FAKE_DHCP_IP_END}],
            'gateway_ip': FAKE_GW_ADDR,
            'ip_version': 4,
            'cidr': FAKE_IP_ADDR + '/24',
            'id': FAKE_SUBNET_ID}}

        dcnmclnt = self.dfa_server.dcnm_client
        dcnmclnt.get_config_profile_for_network.return_value = (
            FAKE_CFG_PROFILE_NAME, FAKE_FWD_MODE)
        self.dfa_server.get_network.return_value = None

        self.dfa_server.network_create_event(network_info)

        fake_network = mock.Mock()
        fake_network.source = 'dcnm'
        fake_network.name = FAKE_NETWORK_NAME
        fake_network.result = constants.SUBNET_PENDING
        self.dfa_server.get_network.return_value = fake_network
        self.dfa_server.subnet_create_event(subnet_info)
        self.assertFalse(self.dfa_server.dcnm_client.create_network.called)

        fake_network.source = 'openstack'
        fake_network.result = constants.SUBNET_PENDING
        self.dfa_server.subnet_create_event(subnet_info)
        self.assertTrue(self.dfa_server.dcnm_client.create_network.called)
        create_call = self.dfa_server.dcnm_client.create_network.call_args
        arg1, arg2 = create_call
        self.assertTrue(arg1[0] == FAKE_PROJECT_NAME)
        self.assertTrue(
            arg1[1].__dict__ == self.dfa_server.network[FAKE_NETWORK_ID2])
        self.assertTrue(
            arg1[2].__dict__ == self.dfa_server.subnet[FAKE_SUBNET_ID])

    def test_network_delete_event(self):
        """Test case for network delete event."""

        self._load_network_info()
        network_info = {'network_id': FAKE_NETWORK_ID}
        self.dfa_server.get_vms.return_value = []
        self.dfa_server.network_delete_event(network_info)
        self.assertTrue(self.dfa_server.dcnm_client.delete_network.called)
        dcall = self.dfa_server.dcnm_client.delete_network.call_args
        arg1, arg2 = dcall
        self.assertTrue(arg1[0] == FAKE_PROJECT_NAME)
        self.assertTrue(arg1[1].name == FAKE_NETWORK_NAME)
        self.assertTrue(arg1[1].segmentation_id == self.segid)
        self.dfa_server.seg_drvr.release_segmentation_id.assert_called_with(
            self.segid)
        self.assertTrue(self.dfa_server.delete_network_db.called)

    def test_dcnm_network_create_event(self):
        """Test case for DCNM network create event."""

        network_info = {'segmentation_id': FAKE_SEG_ID,
                        'project_name': FAKE_PROJECT_NAME,
                        'partition_name': self.part_name}
        self.dfa_server.get_network_by_segid.return_value = None
        self.dfa_server.get_project_id.return_value = FAKE_PROJECT_ID

        dcnm_network = {'segmentId': FAKE_SEG_ID,
                        'profileName': FAKE_CFG_PROFILE_NAME,
                        'networkName': FAKE_NETWORK_NAME,
                        'organizationName': FAKE_PROJECT_NAME,
                        'dhcpScope': None,
                        'netmaskLength': 24,
                        'gateway': FAKE_GW_ADDR}
        self.dfa_server.dcnm_client.get_network.return_value = dcnm_network
        dcnmclnt = self.dfa_server.dcnm_client
        dcnmclnt.config_profile_fwding_mode_get.return_value = FAKE_FWD_MODE

        self.dfa_server.dcnm_network_create_event(network_info)

        # Check the results.
        self.dfa_server.dcnm_client.get_network.assert_called_with(
            FAKE_PROJECT_NAME, FAKE_SEG_ID)
        for netid, dcnmnet in six.iteritems(self.dfa_server.network):
            self.dfa_server.add_network_db.assert_called_with(
                netid, dcnmnet, 'DCNM', constants.RESULT_SUCCESS)
        self.assertTrue(self.dfa_server.neutronclient.create_network.called)
        net_ext_name = self.cfg.dcnm.dcnm_net_ext
        call_args = self.dfa_server.neutronclient.create_network.call_args
        cargs, ckwargs = call_args
        net_name = ckwargs.get('body').get('network').get('name')
        self.assertTrue(net_name == (
            FAKE_NETWORK_NAME + net_ext_name + str(FAKE_SEG_ID)))
        self.assertTrue(self.dfa_server.neutronclient.create_subnet.called)

    def test_dcnm_network_delete_event(self):
        """Test case for DCNM network delete event."""

        self._load_network_info()
        network_info = {'segmentation_id': (
            self.dfa_server.network[FAKE_NETWORK_ID]['segmentation_id'])}

        dcnmnet = mock.Mock()
        dcnmnet.network_id = FAKE_NETWORK_ID
        self.dfa_server.get_network_by_segid.return_value = dcnmnet
        self.dfa_server.dcnm_network_delete_event(network_info)

        # Check the results.
        self.assertTrue(self.dfa_server.network == {})
        self.dfa_server.neutronclient.delete_network.assert_called_with(
            FAKE_NETWORK_ID)
        self.dfa_server.delete_network_db.assert_called_with(FAKE_NETWORK_ID)

    def test_port_create_event(self):
        """Test case for port create event."""

        port_info = self._get_port_info()
        self._load_network_info()
        self.dfa_server._inst_api.get_instance_for_uuid.return_value = (
            FAKE_INSTANCE_NAME)
        self.dfa_server.dcnm_dhcp = True
        self.dfa_server.port_create_event(port_info)

        # Check the output/calls
        self.assertTrue(self.dfa_server.neutron_event.send_vm_info.called)
        call_args = self.dfa_server.neutron_event.send_vm_info.call_args
        cargs, ckwargs = call_args
        self.assertTrue(cargs[0] == FAKE_HOST_ID)
        self.assertTrue(str(self.dfa_server.port[FAKE_PORT_ID]) == cargs[1])
        self.assertTrue(self.dfa_server.add_vms_db.called)
        call_args = self.dfa_server.add_vms_db.call_args
        cargs, ckwargs = call_args
        self.assertTrue(self.dfa_server.port[FAKE_PORT_ID] == cargs[0])
        self.assertTrue(constants.RESULT_SUCCESS == cargs[1])

    def test_port_update_event(self):
        """Test case for port update event."""

        port_info = self._get_port_info()
        mvm = mock.Mock()
        mvm.host = None
        mvm.port_id = FAKE_PORT_ID
        self.dfa_server.get_vm.return_value = mvm
        self.dfa_server._inst_api.get_instance_for_uuid.return_value = (
            FAKE_INSTANCE_NAME)
        self.dfa_server.port_update_event(port_info)

        # Check the results.
        self.dfa_server.neutron_event.send_vm_info.assert_called_with(
            port_info['port']['binding:host_id'],
            str(self.dfa_server.port[port_info['port']['id']]))
        params = dict(columns=dict(
            instance_id=FAKE_DEVICE_ID.replace('-', ''),
            host=port_info['port']['binding:host_id'],
            result=constants.RESULT_SUCCESS,
            name=FAKE_INSTANCE_NAME))
        self.dfa_server.update_vm_db.assert_called_with(
            port_info['port']['id'], **params)

    def test_delete_vm_funciton(self):
        """Test case for port delete event."""

        port_id = FAKE_PORT_ID
        vm = mock.Mock()
        vm.mac = FAKE_MAC_ADDR
        vm.port_id = FAKE_PORT_ID
        vm.segmentation_id = self.segid
        vm.network_id = FAKE_NETWORK_ID,
        vm.port_id = FAKE_PORT_ID
        vm.ip = FAKE_IP_ADDR
        vm.gw_mac = FAKE_GW_ADDR
        vm.instance_id = FAKE_DEVICE_ID
        vm.fwd_mod = FAKE_FWD_MODE
        vm.host = FAKE_HOST_ID
        vm.name = FAKE_INSTANCE_NAME
        self.dfa_server.get_vm.return_value = vm

        # Check the results.

        # Check the output/calls
        self.dfa_server.delete_vm_function(port_id, vm)
        # Check the results.

        self.dfa_server.delete_vm_db.assert_called_with(vm.port_id)

    def test_send_vm_info(self):
        """Test send_send_vm_info"""
        vm = mock.Mock()
        vm.mac = FAKE_MAC_ADDR
        vm.port_id = FAKE_PORT_ID
        vm.segmentation_id = self.segid
        vm.network_id = FAKE_NETWORK_ID,
        vm.port_id = FAKE_PORT_ID
        vm.ip = FAKE_IP_ADDR
        vm.gw_mac = FAKE_GW_ADDR
        vm.instance_id = FAKE_DEVICE_ID
        vm.fwd_mod = FAKE_FWD_MODE
        vm.host = FAKE_HOST_ID
        vm.name = FAKE_INSTANCE_NAME
        vm_info = dict(status='down', vm_mac=vm.mac,
                       segmentation_id=vm.segmentation_id,
                       host=vm.host, port_uuid=vm.port_id,
                       net_uuid=vm.network_id,
                       oui=dict(ip_addr=vm.ip, vm_name=vm.name,
                                vm_uuid=vm.instance_id, gw_mac=vm.gw_mac,
                                fwd_mod=vm.fwd_mod, oui_id='cisco'))
        return_value = self.dfa_server.send_vm_info(vm_info)
        self.assertTrue(self.dfa_server.neutron_event.send_vm_info.called)
        cargs, ckwargs = self.dfa_server.neutron_event.send_vm_info.call_args
        self.assertEqual(FAKE_HOST_ID, cargs[0])
        self.assertEqual(str(vm_info), cargs[1])
        self.assertEqual(True, return_value)

    def test_add_dhcp_port(self):
        """Test add dhcp port"""
        self.dfa_server.get_vm.return_value = None
        port_info = self._get_port_info().get("port")
        self._load_network_info()
        self.dfa_server._inst_api.get_instance_for_uuid.return_value = (
            FAKE_INSTANCE_NAME)
        self.dfa_server.dcnm_dhcp = False
        self.dfa_server.add_dhcp_port(port_info)

        # Check the output/calls

        self.assertTrue(self.dfa_server.neutron_event.send_vm_info.called)
        call_args = self.dfa_server.neutron_event.send_vm_info.call_args
        cargs, ckwargs = call_args
        self.assertEqual(FAKE_HOST_ID, cargs[0])
        self.assertEqual(str(self.dfa_server.port[FAKE_PORT_ID]), cargs[1])
        self.assertEqual(self.dfa_server.port[FAKE_PORT_ID]["oui"]["vm_name"],
                         "dhcp10010_4")
        self.assertTrue(self.dfa_server.add_vms_db.called)
        call_args = self.dfa_server.add_vms_db.call_args
        cargs, ckwargs = call_args
        self.assertEqual(self.dfa_server.port[FAKE_PORT_ID], cargs[0])
        self.assertEqual(constants.RESULT_SUCCESS, cargs[1])

    def test_correct_dhcp_ports(self):
        """Test case for port delete event."""
        port_info = self._get_port_info().get("port")
        port_info["device_owner"] = "network:dhcp"
        ports_list_data = {"ports": [port_info]}
        self.dfa_server.neutronclient.list_ports.return_value = ports_list_data
        self.dfa_server.neutron_event._clients.get.return_value = True
        self.dfa_server.add_dhcp_port = mock.Mock()
        self.dfa_server.correct_dhcp_ports(FAKE_NETWORK_ID)

        self.assertTrue(self.dfa_server.add_dhcp_port.called)
        call_args = self.dfa_server.add_dhcp_port.call_args
        cargs, ckwargs = call_args
        self.assertEqual(FAKE_PORT_ID, cargs[0].get("id"))

    def test_is_mand_arg_present_true(self):
        """Test the is_mand_arg_present function for True case. """
        intf_dict = {'remote_port_id_mac': '00:11:22:33:44:55',
                     'remote_system_name': 'N6K-1'}
        ret = self.rpcb.is_mand_arg_present(intf_dict)
        self.assertTrue(ret)

    def test_is_mand_arg_present_false(self):
        """Test the is_mand_arg_present function for False case. """
        # Recheck this again, it should be an OR condition instead of and
        # Check if both TLVs are received.
        intf_dict = {}
        ret = self.rpcb.is_mand_arg_present(intf_dict)
        self.assertFalse(ret)

    def test_save_topo_disc_params_exist_mand(self):
        """Test the save_topo_disc_params function for exist, mandatory case.

        This is for the case when config is already present in the DB and
        mandatory TLV's are present in the new config. This is the uppdate
        case.
        """
        host = 'host1'
        interface = 'eth2'
        with mock.patch.object(self.rpcb, 'is_mand_arg_present',
                               return_value=True),\
            mock.patch('oslo_serialization.jsonutils.loads') as jsut,\
            mock.patch.object(self.rpcb.obj.topology_db,
                              'add_update_topology_db') as add_upd_mock,\
            mock.patch.object(self.rpcb.obj.topology_db, 'query_topology_db',
                              return_value=[{}]):
            jsut.return_value = {'host': host, 'protocol_interface': interface}
            self.rpcb.save_topo_disc_params(None, None)
        params = dict(columns={'heartbeat': None, 'host': host,
                               'protocol_interface': interface})
        add_upd_mock.assert_called_with(**params)

    def test_save_topo_disc_params_exist_nomand(self):
        """Test the save_topo_disc_* function for exist, non-mandatory case.

        This is for the case when config is already present in the DB and
        mandatory TLV's are not present in the new config. This is the delete
        case.
        """
        host = 'host1'
        interface = 'eth2'
        with mock.patch.object(self.rpcb, 'is_mand_arg_present',
                               return_value=False),\
            mock.patch('oslo_serialization.jsonutils.loads') as jsut,\
            mock.patch.object(self.rpcb.obj.topology_db,
                              'delete_topology_entry') as del_upd_mock,\
            mock.patch.object(self.rpcb.obj.topology_db, 'query_topology_db',
                              return_value=[{}]):
            jsut.return_value = {'host': host, 'protocol_interface': interface}
            self.rpcb.save_topo_disc_params(None, None)
        params = {'host': host, 'protocol_interface': interface}
        del_upd_mock.assert_called_with(**params)

    def test_save_topo_disc_params_nonexist_mand(self):
        """Test the save_topo_disc_* function for non-exist, mandatory case.

        This is for the case when config is not present in the DB and
        mandatory TLV's are present in the new config. This is the add
        case.
        """
        host = 'host1'
        interface = 'eth2'
        with mock.patch.object(self.rpcb, 'is_mand_arg_present',
                               return_value=True),\
            mock.patch('networking_cisco.apps.saf.common.utils.'
                       'utc_time') as utc_mock,\
            mock.patch('oslo_serialization.jsonutils.loads') as jsut,\
            mock.patch.object(self.rpcb.obj.topology_db,
                              'add_update_topology_db') as add_upd_mock,\
            mock.patch.object(self.rpcb.obj.topology_db, 'query_topology_db',
                              return_value=[]):
            jsut.return_value = {'host': host, 'protocol_interface': interface}
            utc_mock.return_value = 'Jan 1'
            self.rpcb.save_topo_disc_params(None, None)
        params = dict(columns={'created': 'Jan 1', 'heartbeat': 'Jan 1',
                               'host': host, 'protocol_interface': interface})
        add_upd_mock.assert_called_with(**params)

    def test_save_topo_disc_params_nonexist_nonmand(self):
        """Test the save_topo_disc_* function for non-exist, non-mand case.

        This is for the case when config is not present in the DB and
        mandatory TLV's are not present in the new config. This is the no-op
        case.
        """
        host = 'host1'
        interface = 'eth2'
        with mock.patch.object(self.rpcb, 'is_mand_arg_present',
                               return_value=False),\
            mock.patch('networking_cisco.apps.saf.common.utils.'
                       'utc_time') as utc_mock,\
            mock.patch('oslo_serialization.jsonutils.loads') as jsut,\
            mock.patch.object(self.rpcb.obj.topology_db,
                              'add_update_topology_db') as add_upd_mock,\
            mock.patch.object(self.rpcb.obj.topology_db, 'query_topology_db',
                              return_value=[]),\
            mock.patch.object(self.rpcb.obj.topology_db,
                              'delete_topology_entry') as del_upd_mock:
            jsut.return_value = {'host': host, 'protocol_interface': interface}
            utc_mock.return_value = 'Jan 1'
            self.rpcb.save_topo_disc_params(None, None)
        add_upd_mock.assert_not_called()
        del_upd_mock.assert_not_called()

    def test_save_topo_disc_params_none_nonexist_nonmand(self):
        """Test the save_topo_disc_* func for none, non-exist, non-mand case.

        This is for the case when config is not present in the DB and
        mandatory TLV's are not present in the new config. The output returned
        is None. This is the no-op case.
        """
        host = 'host1'
        interface = 'eth2'
        with mock.patch.object(self.rpcb, 'is_mand_arg_present',
                               return_value=False),\
            mock.patch('networking_cisco.apps.saf.common.utils.'
                       'utc_time') as utc_mock,\
            mock.patch('oslo_serialization.jsonutils.loads') as jsut,\
            mock.patch.object(self.rpcb.obj.topology_db,
                              'add_update_topology_db') as add_upd_mock,\
            mock.patch.object(self.rpcb.obj.topology_db, 'query_topology_db',
                              return_value=None),\
            mock.patch.object(self.rpcb.obj.topology_db,
                              'delete_topology_entry') as del_upd_mock:
            jsut.return_value = {'host': host, 'protocol_interface': interface}
            utc_mock.return_value = 'Jan 1'
            self.rpcb.save_topo_disc_params(None, None)
        add_upd_mock.assert_not_called()
        del_upd_mock.assert_not_called()

    def test_add_lbaas_port(self):
        port_info = self._get_port_info()
        port_id = port_info.get('port').get("id")
        lb_id = "1111111111111111"
        self.dfa_server.neutronclient.show_port.return_value = (port_info)
        self.dfa_server.add_lbaas_port(port_id, lb_id)

        # Check the output/calls
        self.assertTrue(self.dfa_server.neutron_event.send_vm_info.called)
        self.assertTrue(self.dfa_server.add_vms_db.called)
        call_args = self.dfa_server.add_vms_db.call_args
        cargs, ckwargs = call_args
        self.assertEqual(self.dfa_server.port[FAKE_PORT_ID], cargs[0])
        self.assertEqual(constants.RESULT_SUCCESS, cargs[1])



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""DFA logging helper module."""


import logging
import logging.handlers as log_hdlr
import os
import sys

# Rotating file size limit.
ONEK = 1024
ONEMEG = (1024 * 1024)
MAX_BYTES = (5 * ONEMEG)
BACKUP_COUNT = 10
DAYS = 'D'
SECONDS = 'S'
MINUTES = 'M'
HOURS = 'H'
MIDNIGHT = 'MIDNIGHT'

LOG_LEVELS = {'debug': logging.DEBUG,
              'info': logging.INFO,
              'warning': logging.WARNING,
              'error': logging.ERROR,
              'critical': logging.CRITICAL}

_loggers = {}


def getLogger(name):
    if name in _loggers:
        return _loggers[name]

    logger = logging.getLogger(name)
    _loggers[name] = logger

    return logger


def setup_logger(project, cfg):

    if _loggers.get(project):
        return

    logger = getLogger(None)

    # Set logging level.
    level = LOG_LEVELS.get(cfg.dfa_log.log_level.lower(), logging.WARNING)
    logger.setLevel(level)

    # Set default log levels for specified modules.
    def_log_levels = cfg.dfa_log.default_log_levels
    for dll in def_log_levels.split(','):
        mod, sep, lvl = dll.partition('=')
        level = LOG_LEVELS.get(lvl.lower(), logging.WARNING)
        logging.getLogger(mod).setLevel(level)

    # Set log file path name.
    log_dir = cfg.dfa_log.log_dir
    log_file = cfg.dfa_log.log_file
    if log_dir and log_file:
        try:
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
        except OSError:
            pass
        else:
            log_file_path = os.path.join(log_dir, log_file)
            handler = log_hdlr.TimedRotatingFileHandler(log_file_path,
                                                        when=DAYS,
                                                        interval=1)
    else:
        handler = logging.StreamHandler(sys.stdout)

    if cfg.dfa_log.use_syslog.lower() == 'true':
        log_fac = cfg.dfa_log.syslog_log_facility
        facility = getattr(log_hdlr.SysLogHandler, log_fac, None)
        handler = log_hdlr.SysLogHandler(address='/dev/log',
                                         facility=facility)

    # Setting log format.
    log_format = cfg.dfa_log.log_format
    date_fowrmat = cfg.dfa_log.log_date_format
    formatter = logging.Formatter(fmt=log_format, datefmt=date_fowrmat)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    _loggers[project] = logger



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import abc
from datetime import timedelta
from operator import itemgetter
import random
import six

from sqlalchemy import func
from sqlalchemy.sql import expression as expr

from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.db.device_manager import hd_models


ROUTER_ROLE_HA_REDUNDANCY = cisco_constants.ROUTER_ROLE_HA_REDUNDANCY
# Maximum allowed minute difference in creation time
# for hosting devices to be considered equal
EQUIVALENCE_TIME_DIFF = 7


@six.add_metaclass(abc.ABCMeta)
class L3RouterHostingDeviceBaseScheduler(object):
    """Slot-aware base scheduler of Neutron routers to hosting devices."""

    def get_candidates(self, plugin, context, r_hd_binding_db):
        """Selection criteria: Hosting devices that...
               ... are based on the template required by router's type
               AND
               ... are administratively up
               AND
               ... are active (i.e., has status HD_ACTIVE)
               AND
               ... are bound to tenant owning router OR is unbound
               AND
               ... are enough slots available to host the router

            Among hosting devices meeting these criteria the device with
            less allocated slots is preferred.
        """
        # SELECT hosting_device_id, created_at, sum(num_allocated)
        # FROM hostingdevices AS hd
        # LEFT OUTER JOIN slotallocations AS sa ON hd.id=sa.hosting_device_id
        # WHERE
        #    hd.template_id='11111111-2222-3333-4444-555555555555' AND
        #    hd.admin_state_up=TRUE AND
        #    hd.status='ACTIVE' AND
        # <<<sharing case:>>>
        #    (hd.tenant_bound IS NULL OR hd.tenant_bound='t10')
        # <<<non-sharing case:>>>
        #    (sa.tenant_bound='t10' OR
        #     (sa.tenant_bound IS NULL AND sa.logical_resource_owner='t10') OR
        #     hd.tenant_bound='t10' OR
        #     (hd.tenant_bound IS NULL AND sa.hosting_device_id IS NULL))
        # GROUP BY hosting_device_id
        # HAVING sum(num_allocated) <= 8 OR sum(num_allocated) IS NULL
        # ORDER BY created_at;
        router = r_hd_binding_db.router
        tenant_id = router['tenant_id']
        router_type_db = r_hd_binding_db.router_type
        template_id = router_type_db['template_id']
        template_db = router_type_db.template
        slot_threshold = template_db.slot_capacity - router_type_db.slot_need

        query = context.session.query(
            hd_models.HostingDevice.id, hd_models.HostingDevice.created_at,
            func.sum(hd_models.SlotAllocation.num_allocated))
        query = query.outerjoin(
            hd_models.SlotAllocation,
            hd_models.HostingDevice.id ==
            hd_models.SlotAllocation.hosting_device_id)
        query = query.filter(
            hd_models.HostingDevice.template_id == template_id,
            hd_models.HostingDevice.admin_state_up == expr.true(),
            hd_models.HostingDevice.status == cisco_constants.HD_ACTIVE)
        if r_hd_binding_db.share_hosting_device:
            query = query.filter(
                expr.or_(hd_models.HostingDevice.tenant_bound == expr.null(),
                         hd_models.HostingDevice.tenant_bound == tenant_id))
        else:
            query = query.filter(
                expr.or_(
                    hd_models.SlotAllocation.tenant_bound == tenant_id,
                    expr.and_(
                        hd_models.SlotAllocation.tenant_bound == expr.null(),
                        hd_models.SlotAllocation.logical_resource_owner ==
                        tenant_id),
                    hd_models.HostingDevice.tenant_bound == tenant_id,
                    expr.and_(
                        hd_models.HostingDevice.tenant_bound == expr.null(),
                        hd_models.SlotAllocation.hosting_device_id ==
                        expr.null())))
        query = query.group_by(hd_models.HostingDevice.id)
        query = query.having(
            expr.or_(func.sum(
                hd_models.SlotAllocation.num_allocated) <= slot_threshold,
                func.sum(hd_models.SlotAllocation.num_allocated ==
                         expr.null())))
        query = query.order_by(hd_models.HostingDevice.created_at)
        return query.all()

    @abc.abstractmethod
    def schedule_router(self, plugin, context, r_hd_binding_db):
        # As this is a base scheduler that is not supposed to be used
        # directly we make this method abstract
        pass

    @abc.abstractmethod
    def unschedule_router(self, plugin, context, r_hd_binding_db):
        # As this is a base scheduler that is not supposed to be used
        # directly we make this method abstract
        pass


class L3RouterHostingDeviceLongestRunningScheduler(
        L3RouterHostingDeviceBaseScheduler):
    """Schedules a Neutron router on a hosting device.

        Selection criteria:
            The longest running hosting device that
            has enough slots available to host the router

            Hosting devices with creation date/time less than
            EQUIVALENCE_TIME_DIFF are considered equally old.

            Among hosting devices meeting these criteria and
            that are of same age the device with less allocated
            slots is preferred.
    """
    def schedule_router(self, plugin, context, r_hd_binding_db):
        candidates = self._filtered_candidates(plugin, context,
                                               r_hd_binding_db)
        if len(candidates) == 0:
            # report unsuccessful scheduling
            return
        # determine oldest candidates considered equally old
        oldest_candidates = []
        minute_limit = timedelta(minutes=EQUIVALENCE_TIME_DIFF)
        for candidate in candidates:
            if candidate[1] - candidates[0][1] < minute_limit:
                oldest_candidates.append(candidate)
            else:
                # we're only interested in the longest running devices
                break
        # sort on least number of used slots
        sorted_candidates = sorted(oldest_candidates, key=itemgetter(2))
        return sorted_candidates[0]

    def unschedule_router(self, plugin, context, r_hd_binding_db):
        return True

    def _filtered_candidates(self, plugin, context, r_hd_binding_db):
        return self.get_candidates(plugin, context, r_hd_binding_db)


class CandidatesHAFilterMixin(object):

    def _filtered_candidates(self, plugin, context, r_hd_binding_db):
        candidates_dict = {c.id: c for c in self.get_candidates(
            plugin, context, r_hd_binding_db)}
        if candidates_dict:
            r_b = r_hd_binding_db.router.redundancy_binding
            if r_b:
                # this is a redundancy router so we need to exclude the hosting
                # devices of the user visible router and the other redundancy
                # routers
                if r_b.user_router.hosting_info.hosting_device_id:
                    del candidates_dict[
                        r_b.user_router.hosting_info.hosting_device_id]
                for rr_b in r_b.user_router.redundancy_bindings:
                    rr = rr_b.redundancy_router
                    if (rr.id != r_b.redundancy_router_id and
                            rr.hosting_info.hosting_device_id):
                        del candidates_dict[rr.hosting_info.hosting_device_id]
            elif r_hd_binding_db.role == ROUTER_ROLE_HA_REDUNDANCY:
                # redundancy binding has not been persisted to db yet
                # so we abort this scheduling attempt and retry later
                return []
            for rr_b in r_hd_binding_db.router.redundancy_bindings:
                # this is a user visible router so we need to exclude the
                # hosting devices of its redundancy routers
                rr = rr_b.redundancy_router
                if rr.hosting_info.hosting_device_id:
                    del candidates_dict[rr.hosting_info.hosting_device_id]
        return candidates_dict.values()


class L3RouterHostingDeviceHALongestRunningScheduler(
        CandidatesHAFilterMixin, L3RouterHostingDeviceLongestRunningScheduler):
    """Schedules a Neutron router on a hosting device.

        The scheduler is HA aware and will ignore hosting device candidates
        that are used by other Neutron routers in the same HA group.

        Selection criteria:
            The longest running hosting device that is not already hosting a
            router in the HA group and which has enough slots available to
            host the router.

            Hosting devices with creation date/time less than
            EQUIVALENCE_TIME_DIFF are considered equally old.

            Among hosting devices meeting these criteria and
            that are of same age the device with less allocated
            slots is preferred.
    """
    pass


class L3RouterHostingDeviceRandomScheduler(L3RouterHostingDeviceBaseScheduler):
    """Schedules a Neutron router on a hosting device.

        Selection criteria:
            A randomly selected hosting device that has enough slots available
            to host the router.
    """
    def schedule_router(self, plugin, context, r_hd_binding_db):
        candidates = self._filtered_candidates(plugin, context,
                                               r_hd_binding_db)
        if len(candidates) == 0:
            # report unsuccessful scheduling
            return
        return random.choice(list(candidates))

    def unschedule_router(self, plugin, context, r_hd_binding_db):
        return True

    def _filtered_candidates(self, plugin, context, r_hd_binding):
        return self.get_candidates(plugin, context, r_hd_binding)


class L3RouterHostingDeviceHARandomScheduler(
        CandidatesHAFilterMixin, L3RouterHostingDeviceRandomScheduler):
    """Schedules a Neutron router on a hosting device.

        The scheduler is HA aware and will ignore hosting device candidates
        that are used by other Neutron routers in the same HA group.

        Selection criteria:
            A randomly selected hosting device that is not already hosting a
            router in the HA group and which has enough slots available to
            host the router.
    """
    pass



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock
import unittest

from networking_cisco.plugins.cisco.cpnr.netns import iflist
from networking_cisco.plugins.cisco.cpnr.netns import nslist


class TestNetNs(unittest.TestCase):

    @mock.patch('os.path')
    @mock.patch('os.listdir')
    def test_nslist(self, mock_listdir, mock_path):
        mock_path.exists.return_value = True
        mock_listdir.return_value = []
        mock_listdir.return_value.append(('qdhcp-1111111-2222222-3333333'))
        mock_listdir.return_value.append(('qdhcp-4444444-5555555-6666666'))
        nsdirs = nslist()
        self.assertEqual('qdhcp-1111111-2222222-3333333', nsdirs[0])
        self.assertEqual('qdhcp-4444444-5555555-6666666', nsdirs[1])

        mock_path.exists.return_value = False
        nsdirs = nslist()
        self.assertEqual([], nsdirs)

    @mock.patch('subprocess.check_output')
    def test_iflist(self, mock_check_output):
        ip_addr_str = (
            b'1: lo: <LOOPBACK,UP,LOWER_UP> mtu 16436 qdisc noqueue\n'
            b'  link/loopback 00:00:00:00:00:00 brd 00:00:00:00:00:00\n'
            b'  inet 127.0.0.1/8 scope host lo\n'
            b'  inet6 ::1/128 scope host\n'
            b'      valid_lft forever preferred_lft forever\n '
            b'2: eth0: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500'
            b' qdisc pfifo_fast qlen 1000\n '
            b'  link/ether 00:00:00:00:00:00 brd ff:ff:ff:ff:ff:ff\n'
            b'  inet 10.1.1.1/24 brd 10.1.1.255 scope global eth0\n'
            b'  inet6 1111:2222:3333:4444:5555/64 scope link\n'
            b'      valid_lft forever preferred_lft forever\n'
            b'3: sit0: <NOARP> mtu 1480 qdisc noop\n'
            b'  link/sit 0.0.0.0 brd 0.0.0.0')

        mock_check_output.return_value = ip_addr_str
        interfaces = iflist()
        name, addr, mask = interfaces[0]
        self.assertEqual(b'lo', name)
        self.assertEqual(b'127.0.0.1', addr)
        self.assertEqual(b'8', mask)
        name, addr, mask = interfaces[1]
        self.assertEqual(b'eth0', name)
        self.assertEqual(b'10.1.1.1', addr)
        self.assertEqual(b'24', mask)

        # check ignore option
        interfaces = iflist(ignore=(b"lo",))
        name, addr, mask = interfaces[0]
        self.assertEqual(b'eth0', name)
        self.assertEqual(b'10.1.1.1', addr)
        self.assertEqual(b'24', mask)
        with self.assertRaises(IndexError):
            name, addr, mask = interfaces[1]

        interfaces = iflist(ignore=(b"eth0",))
        name, addr, mask = interfaces[0]
        self.assertEqual(b'lo', name)
        self.assertEqual(b'127.0.0.1', addr)
        self.assertEqual(b'8', mask)
        with self.assertRaises(IndexError):
            name, addr, mask = interfaces[1]

        # test with no input
        mock_check_output.return_value = ''
        interfaces = iflist()
        with self.assertRaises(IndexError):
            name, addr, mask = interfaces[0]



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import subprocess
import sys


def main():
    return subprocess.check_output(['sudo', 'cpnr-rootwrap',
        '/etc/cpnr/rootwrap.conf',
        'cpnr-dhcp-relay-agent'] + sys.argv[1:])


if __name__ == "__main__":
    main()



# Copyright 2015-2016 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import defaultdict
import six
from threading import Timer

from contextlib import contextmanager
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from networking_cisco import backwards_compatibility as bc
from networking_cisco.ml2_drivers.ucsm import config
from networking_cisco.ml2_drivers.ucsm import constants as const
from networking_cisco.ml2_drivers.ucsm import exceptions as cexc
from networking_cisco.ml2_drivers.ucsm import ucsm_db


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class CiscoUcsmDriver(object):

    """UCS Manager Driver Main Class."""

    def __init__(self):
        LOG.debug("UCS Manager Network driver found")
        self.ucsmsdk = None
        self.supported_sriov_vnic_types = [bc.portbindings.VNIC_DIRECT,
                                           bc.portbindings.VNIC_MACVTAP]
        self.supported_pci_devs = CONF.ml2_cisco_ucsm.supported_pci_devs

        config.load_single_ucsm_config()

        self.ucsm_db = ucsm_db.UcsmDbModel()
        self.ucsm_host_dict = {}
        self.ucsm_sp_dict = {}
        self._create_host_and_sp_dicts_from_config()

        # Start timer to monitor Port Profiles that need to be deleted.
        Timer(const.DEFAULT_PP_DELETE_TIME,
            self._delayed_delete_port_profile, ()).start()
        LOG.debug('Starting periodic Port Profile delete timer for %d',
            const.DEFAULT_PP_DELETE_TIME)

    def check_vnic_type_and_vendor_info(self, vnic_type, profile):
        """Checks if this vnic_type and vendor device info are supported.

        Returns True if:
        1. the port vnic_type is direct or macvtap and
        2. the vendor_id and product_id of the port is supported by
        this MD
        Useful in determining if this MD should bind the current
        port.
        """
        # Check for vnic_type
        if vnic_type not in self.supported_sriov_vnic_types:
            LOG.info('Non SR-IOV vnic_type: %s.', vnic_type)
            return False

        if not profile:
            return False

        # Check for vendor_info
        return self._check_for_supported_vendor(profile)

    def _check_for_supported_vendor(self, profile):
        """Checks if the port belongs to a supported vendor.

        Returns True for supported_pci_devs.
        """
        vendor_info = profile.get('pci_vendor_info')
        if not vendor_info:
            return False
        if vendor_info not in self.supported_pci_devs:
            return False
        return True

    def is_vmfex_port(self, profile):
        """Checks if the port is a VMFEX port.

        Returns True only for port that support VM-FEX.
        It is important to distinguish between the two since Port Profiles
        on the UCS Manager are created only for the VM-FEX ports.
        """
        return profile and (profile.get('pci_vendor_info')
            == const.PCI_INFO_CISCO_VIC_1240)

    def _import_ucsmsdk(self):
        """Imports the Ucsm SDK module.

        This module is not installed as part of the normal Neutron
        distributions. It is imported dynamically in this module so that
        the import can be mocked, allowing unit testing without requiring
        the installation of UcsSdk.

        """
        # Check if SSL certificate checking has been disabled.
        # If so, warn the user before proceeding.
        if not CONF.ml2_cisco_ucsm.ucsm_https_verify:
            LOG.warning(const.SSL_WARNING)

        # Monkey patch the UCS sdk version of urllib2 to disable
        # https verify if required.
        from networking_cisco.ml2_drivers.ucsm import ucs_urllib2
        ucsmsdkhandle = importutils.import_module('UcsSdk.UcsHandle')
        ucsmsdkhandle.urllib2 = ucs_urllib2

        ucsmsdk = importutils.import_module('UcsSdk')
        return ucsmsdk

    def _create_host_and_sp_dicts_from_config(self):
        # Pull Service Profile to Hostname mapping from config if it has been
        # provided
        hosts_provided_in_config = False
        for ip, ucsm in CONF.ml2_cisco_ucsm.ucsms.items():
            for host, sp in (ucsm.ucsm_host_list or {}).items():
                hosts_provided_in_config = True
                self.ucsm_host_dict[host] = ip
                if '/' not in sp:
                    self.ucsm_sp_dict[(ip, host)] = (
                        const.SERVICE_PROFILE_PATH_PREFIX + sp.strip())
                else:
                    self.ucsm_sp_dict[(ip, host)] = sp.strip()

        # If no ucsm host list in the config for any ucsm, then learn the hosts
        # and sps from the UCSMs
        if not hosts_provided_in_config:
            self._create_ucsm_host_to_service_profile_mapping()

        if not self.ucsm_sp_dict:
            LOG.error('UCS Manager network driver failed to get Service '
                      'Profile information for any of its nodes.')

    @contextmanager
    def ucsm_connect_disconnect(self, ucsm_ip):
        handle = self.ucs_manager_connect(ucsm_ip)
        try:
            yield handle
        finally:
            self.ucs_manager_disconnect(handle, ucsm_ip)

    def ucs_manager_connect(self, ucsm_ip):
        """Connects to a UCS Manager."""
        if not self.ucsmsdk:
            self.ucsmsdk = self._import_ucsmsdk()

        ucsm = CONF.ml2_cisco_ucsm.ucsms.get(ucsm_ip)

        if not ucsm or not ucsm.ucsm_username or not ucsm.ucsm_password:
            LOG.error('UCS Manager network driver failed to get login '
                      'credentials for UCSM %s', ucsm_ip)
            return None

        handle = self.ucsmsdk.UcsHandle()
        try:
            handle.Login(ucsm_ip, ucsm.ucsm_username, ucsm.ucsm_password)
        except Exception as e:
            # Raise a Neutron exception. Include a description of
            # the original  exception.
            raise cexc.UcsmConnectFailed(ucsm_ip=ucsm_ip, exc=e)

        return handle

    def _get_server_name(self, handle, service_profile_mo, ucsm_ip):
        """Get the contents of the 'Name' field associated with UCS Server.

        When a valid connection hande to UCS Manager is handed in, the Name
        field associated with a UCS Server is returned.
        """
        try:
            resolved_dest = handle.ConfigResolveDn(service_profile_mo.PnDn)
            server_list = resolved_dest.OutConfig.GetChild()
            if not server_list:
                return ""
            return server_list[0].Name
        except Exception as e:
            # Raise a Neutron exception. Include a description of
            # the original  exception.
            raise cexc.UcsmConfigReadFailed(ucsm_ip=ucsm_ip, exc=e)

    def _create_ucsm_host_to_service_profile_mapping(self):
        """Reads list of Service profiles and finds associated Server."""
        ucsm_ips = list(CONF.ml2_cisco_ucsm.ucsms)
        for ucsm_ip in ucsm_ips:
            with self.ucsm_connect_disconnect(ucsm_ip) as handle:
                try:
                    sp_list_temp = handle.ConfigResolveClass('lsServer', None,
                        inHierarchical=False)
                    if sp_list_temp and sp_list_temp.OutConfigs is not None:
                        sp_list = sp_list_temp.OutConfigs.GetChild() or []
                        for sp in sp_list:
                            if sp.PnDn:
                                server_name = self._get_server_name(handle, sp,
                                    ucsm_ip)
                                if (server_name and not sp.OperSrcTemplName):
                                    LOG.debug('Server %s info retrieved '
                                        'from UCSM %s', server_name, ucsm_ip)
                                    key = (ucsm_ip, server_name)
                                    self.ucsm_sp_dict[key] = str(sp.Dn)
                                    self.ucsm_host_dict[server_name] = ucsm_ip
                except Exception as e:
                    # Raise a Neutron exception. Include a description of
                    # the original exception.
                    raise cexc.UcsmConfigReadFailed(ucsm_ip=ucsm_ip, exc=e)

    def _learn_sp_and_template_for_host(self, host_id):
        ucsm_ips = list(CONF.ml2_cisco_ucsm.ucsms)
        for ucsm_ip in ucsm_ips:
            with self.ucsm_connect_disconnect(ucsm_ip) as handle:
                try:
                    sp_list_temp = handle.ConfigResolveClass('lsServer', None,
                        inHierarchical=False)
                    if sp_list_temp and sp_list_temp.OutConfigs is not None:
                        sp_list = sp_list_temp.OutConfigs.GetChild() or []
                        for sp in sp_list:
                            if sp.PnDn:
                                server_name = self._get_server_name(handle, sp,
                                    ucsm_ip)
                                LOG.debug('Server_name = %s', server_name)
                                if server_name == host_id:
                                    self.ucsm_host_dict[server_name] = ucsm_ip
                                    LOG.debug('Found host %s with SP %s '
                                              'at UCSM %s', server_name,
                                              str(sp.Dn), ucsm_ip)
                                    LOG.debug('Server SP Template %s',
                                        sp.OperSrcTemplName)
                                    config.update_sp_template_config(
                                        host_id, ucsm_ip, sp.OperSrcTemplName)
                                    return ucsm_ip
                except Exception as e:
                    # Raise a Neutron exception. Include a description of
                    # the original  exception.
                    raise cexc.UcsmConfigReadFailed(ucsm_ip=ucsm_ip, exc=e)

    def get_ucsm_ip_for_host(self, host_id):
        ucsm_ip = self.ucsm_host_dict.get(host_id)
        if not ucsm_ip:
            # See if there is a UCSM in the configuration with this host_id
            ucsms = CONF.ml2_cisco_ucsm.ucsms
            ucsm_ip = next(iter([ip for ip, ucsm in ucsms.items()
                           if host_id in ucsm.sp_template_list]), None)
            if not ucsm_ip:
                # Try to discover the Service Proile or SP Template for the
                # host directly from UCS Manager
                LOG.debug('Did not find SP Template so reading from UCSM')
                ucsm_ip = self._learn_sp_and_template_for_host(host_id)

        return ucsm_ip

    def _create_vlanprofile(self, handle, vlan_id, ucsm_ip):
        """Creates VLAN profile to be assosiated with the Port Profile."""
        vlan_name = self.make_vlan_name(vlan_id)
        vlan_profile_dest = (const.VLAN_PATH + const.VLAN_PROFILE_PATH_PREFIX +
                             vlan_name)

        try:
            handle.StartTransaction()
            vp1 = handle.GetManagedObject(
                None,
                self.ucsmsdk.FabricLanCloud.ClassId(),
                {self.ucsmsdk.FabricLanCloud.DN: const.VLAN_PATH})
            if not vp1:
                LOG.warning('UCS Manager network driver Vlan Profile '
                            'path at %s missing', const.VLAN_PATH)
                return False

            # Create a vlan profile with the given vlan_id
            vp2 = handle.AddManagedObject(
                vp1,
                self.ucsmsdk.FabricVlan.ClassId(),
                {self.ucsmsdk.FabricVlan.COMPRESSION_TYPE:
                 const.VLAN_COMPRESSION_TYPE,
                 self.ucsmsdk.FabricVlan.DN: vlan_profile_dest,
                 self.ucsmsdk.FabricVlan.SHARING: const.NONE,
                 self.ucsmsdk.FabricVlan.PUB_NW_NAME: "",
                 self.ucsmsdk.FabricVlan.ID: str(vlan_id),
                 self.ucsmsdk.FabricVlan.MCAST_POLICY_NAME: "",
                 self.ucsmsdk.FabricVlan.NAME: vlan_name,
                 self.ucsmsdk.FabricVlan.DEFAULT_NET: "no"})

            handle.CompleteTransaction()
            if vp2:
                LOG.debug('UCS Manager network driver Created Vlan '
                          'Profile %s at %s', vlan_name, vlan_profile_dest)
                return True

        except Exception as e:
            return self._handle_ucsm_exception(e, 'Vlan Profile',
                                               vlan_name, ucsm_ip)

    def _create_port_profile(self, handle, profile_name, vlan_id,
                             vnic_type, ucsm_ip, trunk_vlans, qos_policy):
        """Creates a Port Profile on the UCS Manager.

        Significant parameters set in the port profile are:
        1. Port profile name - Should match what was set in vif_details
        2. High performance mode - For VM-FEX to be enabled/configured on
        the port using this port profile, this mode should be enabled.
        3. Vlan id - Vlan id used by traffic to and from the port.
        """
        port_profile_dest = (const.PORT_PROFILESETDN + const.VNIC_PATH_PREFIX +
                             profile_name)

        vlan_name = self.make_vlan_name(vlan_id)
        vlan_associate_path = (const.PORT_PROFILESETDN +
                               const.VNIC_PATH_PREFIX + profile_name +
                               const.VLAN_PATH_PREFIX + vlan_name)
        cl_profile_name = const.CLIENT_PROFILE_NAME_PREFIX + str(vlan_id)
        cl_profile_dest = (const.PORT_PROFILESETDN + const.VNIC_PATH_PREFIX +
                           profile_name + const.CLIENT_PROFILE_PATH_PREFIX +
                           cl_profile_name)

        # Remove this Port Profile from the delete DB table if it was
        # addded there due to a previous delete.
        self.ucsm_db.remove_port_profile_to_delete(profile_name, ucsm_ip)

        # Check if direct or macvtap mode
        if vnic_type == bc.portbindings.VNIC_DIRECT:
            port_mode = const.HIGH_PERF
        else:
            port_mode = const.NONE

        try:
            handle.StartTransaction()
            port_profile = handle.GetManagedObject(
                None,
                self.ucsmsdk.VnicProfileSet.ClassId(),
                {self.ucsmsdk.VnicProfileSet.DN: const.PORT_PROFILESETDN})

            if not port_profile:
                LOG.warning('UCS Manager network driver Port Profile '
                            'path at %s missing',
                    const.PORT_PROFILESETDN)
                return False

            # Create a port profile on the UCS Manager
            p_profile = handle.AddManagedObject(
                port_profile,
                self.ucsmsdk.VnicProfile.ClassId(),
                {self.ucsmsdk.VnicProfile.NAME: profile_name,
                 self.ucsmsdk.VnicProfile.POLICY_OWNER: "local",
                 self.ucsmsdk.VnicProfile.NW_CTRL_POLICY_NAME: "",
                 self.ucsmsdk.VnicProfile.PIN_TO_GROUP_NAME: "",
                 self.ucsmsdk.VnicProfile.DN: port_profile_dest,
                 self.ucsmsdk.VnicProfile.DESCR: const.DESCR,
                 self.ucsmsdk.VnicProfile.QOS_POLICY_NAME: qos_policy,
                 self.ucsmsdk.VnicProfile.HOST_NW_IOPERF: port_mode,
                 self.ucsmsdk.VnicProfile.MAX_PORTS: const.MAX_PORTS})
            if not p_profile:
                LOG.warning('UCS Manager network driver could not '
                            'create Port Profile %s.', profile_name)
                return False

            LOG.debug('UCS Manager network driver associating Vlan '
                      'Profile with Port Profile at %s',
                vlan_associate_path)
            # Associate port profile with vlan profile
            mo = handle.AddManagedObject(
                p_profile,
                self.ucsmsdk.VnicEtherIf.ClassId(),
                {self.ucsmsdk.VnicEtherIf.DN: vlan_associate_path,
                 self.ucsmsdk.VnicEtherIf.NAME: vlan_name,
                 self.ucsmsdk.VnicEtherIf.DEFAULT_NET: "yes"}, True)
            if not mo:
                LOG.warning('UCS Manager network driver cannot '
                            'associate Vlan Profile to Port '
                            'Profile %s', profile_name)
                return False
            LOG.debug('UCS Manager network driver created Port Profile %s '
                      'at %s', profile_name, port_profile_dest)

            # For Multi VLAN trunk support
            if trunk_vlans:
                for vlan in trunk_vlans:
                    vlan_name = self.make_vlan_name(vlan)
                    vlan_associate_path = (const.PORT_PROFILESETDN +
                        const.VNIC_PATH_PREFIX + profile_name +
                        const.VLAN_PATH_PREFIX + vlan_name)
                    # Associate port profile with vlan profile
                    # for the trunk vlans
                    mo = handle.AddManagedObject(
                        p_profile,
                        self.ucsmsdk.VnicEtherIf.ClassId(),
                        {self.ucsmsdk.VnicEtherIf.DN: vlan_associate_path,
                        self.ucsmsdk.VnicEtherIf.NAME: vlan_name,
                        self.ucsmsdk.VnicEtherIf.DEFAULT_NET: "no"}, True)
                    if not mo:
                        LOG.warning('UCS Manager network driver cannot '
                                    'associate Vlan %(vlan)d to Port '
                                    'Profile %(profile)s',
                                    {'vlan': vlan, 'profile': profile_name})

            cl_profile = handle.AddManagedObject(
                p_profile,
                self.ucsmsdk.VmVnicProfCl.ClassId(),
                {self.ucsmsdk.VmVnicProfCl.ORG_PATH: ".*",
                 self.ucsmsdk.VmVnicProfCl.DN: cl_profile_dest,
                 self.ucsmsdk.VmVnicProfCl.NAME: cl_profile_name,
                 self.ucsmsdk.VmVnicProfCl.POLICY_OWNER: "local",
                 self.ucsmsdk.VmVnicProfCl.SW_NAME: ".*",
                 self.ucsmsdk.VmVnicProfCl.DC_NAME: ".*",
                 self.ucsmsdk.VmVnicProfCl.DESCR: const.DESCR})
            handle.CompleteTransaction()

            if not cl_profile:
                LOG.warning('UCS Manager network driver could not '
                            'create Client Profile %s.',
                            cl_profile_name)
                return False

            LOG.debug('UCS Manager network driver created Client Profile '
                      '%s at %s', cl_profile_name, cl_profile_dest)
            return True

        except Exception as e:
            return self._handle_ucsm_exception(e, 'Port Profile',
                                               profile_name, ucsm_ip)

    def create_portprofile(self, profile_name, vlan_id, vnic_type, host_id,
        trunk_vlans):
        """Top level method to create Port Profiles on the UCS Manager.

        Calls all the methods responsible for the individual tasks that
        ultimately result in the creation of the Port Profile on the UCS
        Manager.
        """
        ucsm_ip = self.get_ucsm_ip_for_host(host_id)
        if not ucsm_ip:
            LOG.info('UCS Manager network driver does not have UCSM IP '
                     'for Host_id %s', str(host_id))
            return False

        with self.ucsm_connect_disconnect(ucsm_ip) as handle:
            # Create Vlan Profile
            if not self._create_vlanprofile(handle, vlan_id, ucsm_ip):
                LOG.error('UCS Manager network driver failed to create '
                          'Vlan Profile for vlan %s', str(vlan_id))
                return False
            if trunk_vlans:
                for vlan in trunk_vlans:
                    if not self._create_vlanprofile(handle, vlan, ucsm_ip):
                        LOG.error('UCS Manager network driver failed to '
                                  'create Vlan Profile for vlan %s', vlan)
                        return False

            qos_policy = CONF.ml2_cisco_ucsm.ucsms[ucsm_ip].sriov_qos_policy
            if qos_policy:
                LOG.debug('UCS Manager Network driver applying QoS Policy '
                          '%(qos)s to Port Profile %(port_profile)s',
                          {'qos': qos_policy, 'port_profile': profile_name})

            # Create Port Profile
            if not self._create_port_profile(handle, profile_name,
                                             vlan_id, vnic_type,
                                             ucsm_ip, trunk_vlans,
                                             qos_policy):
                LOG.error('UCS Manager network driver failed to create '
                          'Port Profile %s', profile_name)
                return False

        return True

    def _update_service_profile(self, handle, service_profile,
                                vlan_id, ucsm_ip):
        """Updates Service Profile on the UCS Manager.

        Each of the ethernet ports on the Service Profile representing
        the UCS Server, is updated with the VLAN profile corresponding
        to the vlan_id passed in.
        """
        virtio_port_list = (
            CONF.ml2_cisco_ucsm.ucsms[ucsm_ip].ucsm_virtio_eth_ports)

        eth_port_paths = ["%s%s" % (service_profile, ep)
            for ep in virtio_port_list]

        vlan_name = self.make_vlan_name(vlan_id)

        try:
            handle.StartTransaction()
            obj = handle.GetManagedObject(
                None,
                self.ucsmsdk.LsServer.ClassId(),
                {self.ucsmsdk.LsServer.DN: service_profile})

            if not obj:
                LOG.debug('UCS Manager network driver could not find '
                          'Service Profile %s in UCSM %s',
                          service_profile, ucsm_ip)
                return False

            for eth_port_path in eth_port_paths:
                eth = handle.GetManagedObject(
                    obj, self.ucsmsdk.VnicEther.ClassId(),
                    {self.ucsmsdk.VnicEther.DN: eth_port_path}, True)

                if eth:
                    vlan_path = (eth_port_path + const.VLAN_PATH_PREFIX +
                                 vlan_name)

                    eth_if = handle.AddManagedObject(eth,
                        self.ucsmsdk.VnicEtherIf.ClassId(),
                        {self.ucsmsdk.VnicEtherIf.DN: vlan_path,
                        self.ucsmsdk.VnicEtherIf.NAME: vlan_name,
                        self.ucsmsdk.VnicEtherIf.DEFAULT_NET: "no"}, True)

                    if not eth_if:
                        LOG.debug('UCS Manager network driver could not '
                                  'update Service Profile %s with vlan %d',
                                  service_profile, vlan_id)
                        return False
                else:
                    LOG.debug('UCS Manager network driver did not find '
                              'ethernet port at %s', eth_port_path)

            handle.CompleteTransaction()
            return True

        except Exception as e:
            return self._handle_ucsm_exception(e, 'Service Profile',
                                               vlan_name, ucsm_ip)

    def update_service_profile_template(self, vlan_id,
                                        host_id, ucsm_ip):

        template_info = (
            CONF.ml2_cisco_ucsm.ucsms[ucsm_ip].sp_template_list[host_id])

        sp_template_path = (template_info.path + const.SP_TEMPLATE_PREFIX +
                            template_info.name)

        vlan_name = self.make_vlan_name(vlan_id)

        with self.ucsm_connect_disconnect(ucsm_ip) as handle:
            # Create Vlan Profile
            if not self._create_vlanprofile(handle, vlan_id, ucsm_ip):
                LOG.error('UCS Manager network driver failed to create '
                          'Vlan Profile for vlan %s', str(vlan_id))
                return False

            virtio_port_list = (
                CONF.ml2_cisco_ucsm.ucsms[ucsm_ip].ucsm_virtio_eth_ports)
            eth_port_paths = ["%s%s" % (sp_template_path, ep)
                for ep in virtio_port_list]

            try:
                handle.StartTransaction()
                obj = handle.GetManagedObject(
                    None,
                    self.ucsmsdk.LsServer.ClassId(),
                    {self.ucsmsdk.LsServer.DN: sp_template_path})
                if not obj:
                    LOG.error('UCS Manager network driver could not find '
                              'Service Profile template %s.',
                        sp_template_path)
                    return False

                for eth_port_path in eth_port_paths:
                    eth = handle.GetManagedObject(
                        obj, self.ucsmsdk.VnicEther.ClassId(),
                        {self.ucsmsdk.VnicEther.DN: eth_port_path}, True)

                    if eth:
                        vlan_path = (eth_port_path + const.VLAN_PATH_PREFIX +
                                     vlan_name)

                        eth_if = handle.AddManagedObject(eth,
                            self.ucsmsdk.VnicEtherIf.ClassId(),
                            {self.ucsmsdk.VnicEtherIf.DN: vlan_path,
                            self.ucsmsdk.VnicEtherIf.NAME: vlan_name,
                            self.ucsmsdk.VnicEtherIf.DEFAULT_NET: "no"}, True)

                        if not eth_if:
                            LOG.debug('UCS Manager network driver could not '
                                      'update Service Profile Template %s '
                                      'with vlan %d',
                                      template_info.name, vlan_id)
                            return False
                    else:
                        LOG.debug('UCS Manager network driver did not find '
                                  'ethernet port at %s', eth_port_path)
                handle.CompleteTransaction()
                return True
            except Exception as e:
                return self._handle_ucsm_exception(e,
                                                   'Service Profile Template',
                                                   vlan_id, ucsm_ip)

    def update_serviceprofile(self, host_id, vlan_id):
        """Top level method to update Service Profiles on UCS Manager.

        Calls all the methods responsible for the individual tasks that
        ultimately result in a vlan_id getting programed on a server's
        ethernet ports and the Fabric Interconnect's network ports.
        """
        ucsm_ip = self.get_ucsm_ip_for_host(host_id)
        if not ucsm_ip:
            LOG.info('UCS Manager network driver does not have UCSM IP '
                     'for Host_id %s', str(host_id))
            return False

        service_profile = self.ucsm_sp_dict.get((ucsm_ip, host_id))
        if service_profile:
            LOG.debug('UCS Manager network driver Service Profile : %s',
                      service_profile)
        else:
            LOG.info('UCS Manager network driver does not support '
                     'Host_id %s', host_id)
            return False

        with self.ucsm_connect_disconnect(ucsm_ip) as handle:
            # Create Vlan Profile
            if not self._create_vlanprofile(handle, vlan_id, ucsm_ip):
                LOG.error('UCS Manager network driver failed to create '
                          'Vlan Profile for vlan %s', str(vlan_id))
                return False

            # Update Service Profile
            if not self._update_service_profile(handle,
                                                service_profile,
                                                vlan_id,
                                                ucsm_ip):
                LOG.error('UCS Manager network driver failed to update '
                          'Service Profile %(service_profile)s in UCSM '
                          '%(ucsm_ip)s',
                    {'service_profile': service_profile, 'ucsm_ip': ucsm_ip})
                return False

        return True

    def update_vnic_template(self, host_id, vlan_id, physnet,
        vnic_template_path, vnic_template):
        """Updates VNIC Template with the vlan_id."""
        ucsm_ip = self.get_ucsm_ip_for_host(host_id)

        if not ucsm_ip:
            LOG.info('UCS Manager network driver does not have UCSM IP '
                     'for Host_id %s', str(host_id))
            return False

        vlan_name = self.make_vlan_name(vlan_id)

        with self.ucsm_connect_disconnect(ucsm_ip) as handle:
            # Create Vlan Profile
            if not self._create_vlanprofile(handle, vlan_id, ucsm_ip):
                LOG.error('UCS Manager network driver failed to create '
                          'Vlan Profile for vlan %s', vlan_id)
                return False

            try:
                LOG.debug('VNIC Template Path: %s', vnic_template_path)
                vnic_template_full_path = (vnic_template_path +
                   const.VNIC_TEMPLATE_PREFIX + str(vnic_template))
                LOG.debug('VNIC Template Path: %s for physnet %s',
                    vnic_template_full_path, physnet)

                handle.StartTransaction()
                mo = handle.GetManagedObject(
                    None,
                    self.ucsmsdk.VnicLanConnTempl.ClassId(),
                    {self.ucsmsdk.VnicLanConnTempl.DN:
                    vnic_template_full_path}, True)
                if not mo:
                    LOG.error('UCS Manager network driver could '
                              'not find VNIC template %s',
                        vnic_template_full_path)
                    return False

                vlan_dn = (vnic_template_full_path + const.VLAN_PATH_PREFIX +
                    vlan_name)
                LOG.debug('VNIC Template VLAN path: %s', vlan_dn)
                eth_if = handle.AddManagedObject(mo,
                    self.ucsmsdk.VnicEtherIf.ClassId(),
                    {self.ucsmsdk.VnicEtherIf.DN: vlan_dn,
                    self.ucsmsdk.VnicEtherIf.NAME: vlan_name,
                    self.ucsmsdk.VnicEtherIf.DEFAULT_NET: "no"}, True)
                if not eth_if:
                    LOG.error('UCS Manager network driver could '
                              'not add VLAN %(vlan_name)s to VNIC '
                              'template %(vnic_template_full_path)s',
                        {'vlan_name': vlan_name,
                        'vnic_template_full_path': vnic_template_full_path})
                    return False

                handle.CompleteTransaction()
                return True
            except Exception as e:
                return self._handle_ucsm_exception(e, 'VNIC Template',
                    vlan_id, ucsm_ip)

    def _delete_vlan_profile(self, handle, vlan_id, ucsm_ip):
        """Deletes VLAN Profile from UCS Manager."""
        vlan_name = self.make_vlan_name(vlan_id)
        vlan_profile_dest = (const.VLAN_PATH + const.VLAN_PROFILE_PATH_PREFIX +
                             vlan_name)
        try:
            handle.StartTransaction()
            obj = handle.GetManagedObject(
                None,
                self.ucsmsdk.FabricVlan.ClassId(),
                {self.ucsmsdk.FabricVlan.DN: vlan_profile_dest})

            if obj:
                handle.RemoveManagedObject(obj)

            handle.CompleteTransaction()

        except Exception as e:
            # Raise a Neutron exception. Include a description of
            # the original  exception.
            raise cexc.UcsmConfigFailed(config=vlan_id,
                                        ucsm_ip=ucsm_ip, exc=e)

    def _delayed_delete_port_profile(self):
        pp_delete_dict = defaultdict(list)
        Timer(const.DEFAULT_PP_DELETE_TIME,
            self._delayed_delete_port_profile, ()).start()
        all_pps = self.ucsm_db.get_all_port_profiles_to_delete()
        for pp in all_pps:
            pp_delete_dict[pp.device_id].append(pp.profile_id)

        # Connect to each UCSM IP and try to delete Port profiles
        for ucsm_ip in pp_delete_dict.keys():
            with self.ucsm_connect_disconnect(ucsm_ip) as handle:
                for pp in pp_delete_dict.get(ucsm_ip):
                    LOG.debug('Deleting PP %s from UCSM %s', pp,
                        ucsm_ip)
                    try:
                        self._delete_port_profile_from_ucsm(handle,
                                                            pp, ucsm_ip)
                        # Remove this Port Profile from the delete DB table
                        # if it was addded there due to a previous delete.
                        LOG.debug('Removing PP %s from delete table after '
                                  'successful delete', pp)
                        self.ucsm_db.remove_port_profile_to_delete(pp, ucsm_ip)
                    except Exception:
                        #do nothing
                        LOG.debug('Could not delete PP %s from UCSM %s',
                            pp, ucsm_ip)

    def _delete_port_profile_from_ucsm(self, handle, port_profile, ucsm_ip):
        """Deletes Port Profile from UCS Manager."""
        port_profile_dest = (const.PORT_PROFILESETDN + const.VNIC_PATH_PREFIX +
                             port_profile)
        handle.StartTransaction()

        # Find port profile on the UCS Manager
        p_profile = handle.GetManagedObject(
            None,
            self.ucsmsdk.VnicProfile.ClassId(),
            {self.ucsmsdk.VnicProfile.NAME: port_profile,
            self.ucsmsdk.VnicProfile.DN: port_profile_dest})

        if p_profile:
            handle.RemoveManagedObject(p_profile)
        else:
            LOG.warning('UCS Manager network driver did not find '
                        'Port Profile %s to delete.',
                        port_profile)

        handle.CompleteTransaction()

    def _delete_port_profile(self, handle, port_profile, ucsm_ip):
        """Calls method to delete Port Profile from UCS Manager.
           If exception is raised by UCSM, then the PP is added to
           a DB table. The delete timer thread, tried to delete all
           PPs added to this table when it wakes up.
           """
        try:
            self._delete_port_profile_from_ucsm(handle, port_profile, ucsm_ip)

        except Exception as e:
            # Add the Port Profile that we could not delete to the Port Profile
            # delete table. A periodic task will attempt to delete it.
            LOG.debug('Received Port Profile delete exception %s', e)
            self.ucsm_db.add_port_profile_to_delete_table(port_profile,
                                                          ucsm_ip)

    def _remove_vlan_from_all_service_profiles(self, handle, vlan_id, ucsm_ip):
        """Deletes VLAN Profile config from server's ethernet ports."""
        service_profile_list = []
        for key, value in six.iteritems(self.ucsm_sp_dict):
            if (ucsm_ip in key) and value:
                service_profile_list.append(value)

        if not service_profile_list:
            # Nothing to do
            return

        try:
            handle.StartTransaction()
            for service_profile in service_profile_list:
                virtio_port_list = (
                    CONF.ml2_cisco_ucsm.ucsms[ucsm_ip].ucsm_virtio_eth_ports)
                eth_port_paths = ["%s%s" % (service_profile, ep)
                    for ep in virtio_port_list]

                # 1. From the Service Profile config, access the
                # configuration for its ports.
                # 2. Check if that Vlan has been configured on each port
                # 3. If Vlan config found, remove it.
                obj = handle.GetManagedObject(
                        None,
                        self.ucsmsdk.LsServer.ClassId(),
                        {self.ucsmsdk.LsServer.DN: service_profile})

                if obj:
                    # Check if this vlan_id has been configured on the
                    # ports in this Service profile
                    for eth_port_path in eth_port_paths:
                        eth = handle.GetManagedObject(
                            obj, self.ucsmsdk.VnicEther.ClassId(),
                            {self.ucsmsdk.VnicEther.DN: eth_port_path},
                            True)
                        if eth:
                            vlan_name = self.make_vlan_name(vlan_id)
                            vlan_path = eth_port_path + "/if-" + vlan_name
                            vlan = handle.GetManagedObject(eth,
                                self.ucsmsdk.VnicEtherIf.ClassId(),
                                {self.ucsmsdk.VnicEtherIf.DN: vlan_path})
                            if vlan:
                                # Found vlan config. Now remove it.
                                handle.RemoveManagedObject(vlan)
            handle.CompleteTransaction()

        except Exception as e:
            # Raise a Neutron exception. Include a description of
            # the original  exception.
            raise cexc.UcsmConfigDeleteFailed(config=vlan_id,
                                              ucsm_ip=ucsm_ip,
                                              exc=e)

    def _remove_vlan_from_all_sp_templates(self, handle, vlan_id, ucsm_ip):
        """Deletes VLAN config from all SP Templates that have it."""
        sp_template_info_list = (
            CONF.ml2_cisco_ucsm.ucsms[ucsm_ip].sp_template_list.values())

        vlan_name = self.make_vlan_name(vlan_id)
        virtio_port_list = (
            CONF.ml2_cisco_ucsm.ucsms[ucsm_ip].ucsm_virtio_eth_ports)

        try:
            handle.StartTransaction()
            # sp_template_info_list is a list of tuples.
            # Each tuple is of the form :
            # (ucsm_ip, sp_template_path, sp_template)
            for sp_template_info in sp_template_info_list:
                sp_template_path = sp_template_info.path
                sp_template = sp_template_info.name

                sp_template_full_path = (sp_template_path +
                    const.SP_TEMPLATE_PREFIX + sp_template)

                obj = handle.GetManagedObject(
                    None,
                    self.ucsmsdk.LsServer.ClassId(),
                    {self.ucsmsdk.LsServer.DN: sp_template_full_path})
                if not obj:
                    LOG.error('UCS Manager network driver could not '
                        'find Service Profile template %s',
                        sp_template_full_path)
                    continue

                eth_port_paths = ["%s%s" % (sp_template_full_path, ep)
                    for ep in virtio_port_list]
                for eth_port_path in eth_port_paths:
                    eth = handle.GetManagedObject(
                        obj, self.ucsmsdk.VnicEther.ClassId(),
                        {self.ucsmsdk.VnicEther.DN: eth_port_path}, True)

                    if eth:
                        vlan_path = (eth_port_path +
                            const.VLAN_PATH_PREFIX + vlan_name)
                        vlan = handle.GetManagedObject(eth,
                            self.ucsmsdk.VnicEtherIf.ClassId(),
                            {self.ucsmsdk.VnicEtherIf.DN: vlan_path})
                        if vlan:
                            # Found vlan config. Now remove it.
                            handle.RemoveManagedObject(vlan)
                        else:
                            LOG.debug('UCS Manager network driver did not '
                            'find VLAN %s at %s', vlan_name, eth_port_path)
                    else:
                        LOG.debug('UCS Manager network driver did not '
                            'find ethernet port at %s', eth_port_path)
                handle.CompleteTransaction()
                return True
        except Exception as e:
            # Raise a Neutron exception. Include a description of
            # the original  exception.
            raise cexc.UcsmConfigDeleteFailed(config=vlan_id,
                                              ucsm_ip=ucsm_ip,
                                              exc=e)

    def _remove_vlan_from_vnic_templates(self, handle, vlan_id, ucsm_ip):
        """Removes VLAN from all VNIC templates that have it enabled."""
        ucsm = CONF.ml2_cisco_ucsm.ucsms[ucsm_ip]
        vnic_template_info = ucsm.vnic_template_list.values()

        vlan_name = self.make_vlan_name(vlan_id)

        if not vnic_template_info:
            # Nothing to do
            return
        try:
            handle.StartTransaction()
            for temp_info in vnic_template_info:
                vnic_template = temp_info.template
                vnic_template_path = temp_info.path

                vnic_template_full_path = (vnic_template_path +
                    const.VNIC_TEMPLATE_PREFIX + str(vnic_template))
                LOG.debug('vnic_template_full_path: %s',
                    vnic_template_full_path)
                mo = handle.GetManagedObject(
                    None,
                    self.ucsmsdk.VnicLanConnTempl.ClassId(),
                    {self.ucsmsdk.VnicLanConnTempl.DN: (
                        vnic_template_full_path)},
                    True)
                if not mo:
                    LOG.error('UCS Manager network driver could '
                              'not find VNIC template %s at',
                        vnic_template_full_path)
                    continue

                vlan_dn = (vnic_template_full_path +
                    const.VLAN_PATH_PREFIX + vlan_name)
                LOG.debug('VNIC Template VLAN path; %s', vlan_dn)
                eth_if = handle.GetManagedObject(mo,
                    self.ucsmsdk.VnicEtherIf.ClassId(),
                    {self.ucsmsdk.VnicEtherIf.DN: vlan_dn})

                if not eth_if:
                    LOG.error('UCS Manager network driver could not '
                              'delete VLAN %(vlan_name)s from VNIC '
                              'template %(vnic_template_full_path)s',
                        {'vlan_name': vlan_name,
                        'vnic_template_full_path':
                        vnic_template_full_path})
                if eth_if:
                    handle.RemoveManagedObject(eth_if)
            handle.CompleteTransaction()
            return True
        except Exception as e:
            return self._handle_ucsm_exception(e, 'VNIC Template',
                                               vlan_id, ucsm_ip)

    def delete_all_config_for_vlan(self, vlan_id, port_profile,
                                   trunk_vlans):
        """Top level method to delete all config for vlan_id."""
        ucsm_ips = list(CONF.ml2_cisco_ucsm.ucsms)
        for ucsm_ip in ucsm_ips:

            with self.ucsm_connect_disconnect(ucsm_ip) as handle:
                LOG.debug('Deleting config for VLAN %d from UCSM %s', vlan_id,
                    ucsm_ip)
                if (port_profile):
                    self._delete_port_profile(handle, port_profile, ucsm_ip)

                ucsm = CONF.ml2_cisco_ucsm.ucsms[ucsm_ip]
                if ucsm.sp_template_list:
                    self._remove_vlan_from_all_sp_templates(handle,
                                                            vlan_id,
                                                            ucsm_ip)
                if ucsm.vnic_template_list:
                    self._remove_vlan_from_vnic_templates(handle,
                                                          vlan_id,
                                                          ucsm_ip)
                if not (ucsm.sp_template_list and
                        ucsm.vnic_template_list):
                    self._remove_vlan_from_all_service_profiles(handle,
                                                                vlan_id,
                                                                ucsm_ip)
                self._delete_vlan_profile(handle, vlan_id, ucsm_ip)
                if trunk_vlans:
                    for vlan_id in trunk_vlans:
                        self._delete_vlan_profile(handle, vlan_id, ucsm_ip)

    def _handle_ucsm_exception(self, exception_type, profile_type,
                               profile_name, ucsm_ip):
        if const.DUPLICATE_EXCEPTION in str(exception_type):
            LOG.debug('UCS Manager network driver ignoring duplicate '
                      'create/update of %s with %s',
                profile_type, profile_name)
            return True
        else:
            # Raise a Neutron exception. Include a description of
            # the original  exception.
            raise cexc.UcsmConfigFailed(config=profile_name,
                                        ucsm_ip=ucsm_ip,
                                        exc=exception_type)

    def ucs_manager_disconnect(self, handle, ucsm_ip):
        """Disconnects from the UCS Manager.

        After the disconnect, the handle associated with this connection
        is no longer valid.
        """
        try:
            handle.Logout()
        except Exception as e:
            # Raise a Neutron exception. Include a description of
            # the original  exception.
            raise cexc.UcsmDisconnectFailed(ucsm_ip=ucsm_ip, exc=e)

    @staticmethod
    def make_vlan_name(vlan_id):
        return const.VLAN_PROFILE_NAME_PREFIX + str(vlan_id)



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import os
import platform
import sys
import time

import eventlet
eventlet.monkey_patch()
from oslo_serialization import jsonutils

from networking_cisco.apps.saf.agent import iptables_driver as iptd
from networking_cisco.apps.saf.agent.vdp import dfa_vdp_mgr as vdpm
from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.common import constants
from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import rpc
from networking_cisco.apps.saf.common import utils


LOG = logging.getLogger(__name__)

thishost = platform.node()


class RpcCallBacks(object):

    """RPC call back methods."""

    def __init__(self, vdpd, ipt_drvr):
        self._vdpd = vdpd
        self._iptd = ipt_drvr

    def _enqueue_event(self, vm):
        oui = vm.get('oui')
        if oui and oui.get('ip_addr') != '0.0.0.0' and self._iptd:
            rule_info = dict(mac=vm.get('vm_mac'),
                             ip=oui.get('ip_addr'),
                             port=vm.get('port_uuid'),
                             status=vm.get('status'))
            self._iptd.enqueue_event(rule_info)

    def send_vm_info(self, context, msg):
        vm_info = eval(msg)
        LOG.debug('Received vm_info: %s', vm_info)
        # Call VDP/LLDPad API to send the info
        self._vdpd.vdp_vm_event(vm_info)

        # Enqueue the vm info for updating iptables.
        if isinstance(vm_info, list):
            for vm in vm_info:
                self._enqueue_event(vm)
        else:
            self._enqueue_event(vm_info)

    def update_ip_rule(self, context, msg):
        rule_info = eval(msg)
        LOG.debug('RX Info : %s', rule_info)
        # Update the iptables for this rule
        if self._iptd:
            self._iptd.enqueue_event(rule_info)

    def send_msg_to_agent(self, context, msg):
        msg_type = context.get('type')
        uplink = jsonutils.loads(msg)
        LOG.debug("Received %(context)s and %(msg)s", (
            {'context': context, 'msg': uplink}))
        if msg_type == constants.UPLINK_NAME:
            LOG.debug("uplink is %(uplink)s", uplink)
            self._vdpd.dfa_uplink_restart(uplink)


class DfaAgent(object):

    """DFA agent."""

    def __init__(self, host, rpc_qn):
        self._host_name = host
        self._cfg = config.CiscoDFAConfig('neutron').cfg
        self._my_host = self._cfg.DEFAULT.host if self._cfg.DEFAULT.host else (
            utils.find_agent_host_id(host))
        self._qn = '_'.join((rpc_qn, self._my_host))
        LOG.debug('Starting DFA Agent on %s', self._my_host)

        # List of task in the agent
        self.agent_task_list = []

        # This flag indicates the agent started for the first time.
        self._need_uplink_info = True

        # Initialize iptables driver. This will be used to update the ip
        # rules in iptables, after launching an instance.

        if (self._cfg.dcnm.dcnm_dhcp.lower() == 'true'):
            self._iptd = iptd.IptablesDriver(self._cfg)
        else:
            self._iptd = None
            LOG.debug("Using native dhcp, iptable driver is not needed")

        # Setup RPC client for sending heartbeat to controller
        self._url = self._cfg.dfa_rpc.transport_url
        self.setup_client_rpc()

        # Initialize VPD manager.
        br_int = self._cfg.dfa_agent.integration_bridge
        br_ext = self._cfg.dfa_agent.external_dfa_bridge
        config_dict = {'integration_bridge': br_int,
                       'external_bridge': br_ext,
                       'host_id': self._my_host,
                       'root_helper': self._cfg.sys.root_helper,
                       'node_list': self._cfg.general.node,
                       'node_uplink_list': self._cfg.general.node_uplink}

        self._vdpm = vdpm.VdpMgr(config_dict, self.clnt, self._host_name)
        self.pool = eventlet.GreenPool()
        self.setup_rpc()

    def setup_client_rpc(self):
        """Setup RPC client for dfa agent."""
        # Setup RPC client.
        self.clnt = rpc.DfaRpcClient(self._url, constants.DFA_SERVER_QUEUE,
                                     exchange=constants.DFA_EXCHANGE)

    def send_heartbeat(self):
        context = {}
        args = jsonutils.dumps(dict(when=time.ctime(), agent=thishost))
        msg = self.clnt.make_msg('heartbeat', context, msg=args)
        resp = self.clnt.cast(msg)
        LOG.debug("send_heartbeat: resp = %s", resp)

    def request_uplink_info(self):
        context = {}
        msg = self.clnt.make_msg('request_uplink_info',
                                 context, agent=self._my_host)
        try:
            resp = self.clnt.call(msg)
            LOG.debug("request_uplink_info: resp = %s", resp)
            self._need_uplink_info = resp
        except rpc.MessagingTimeout:
            LOG.error("RPC timeout: Request for uplink info failed.")

    def is_uplink_received(self):
        """Finds if uplink information is received and processed. """

        return self._vdpm.is_uplink_received()

    def setup_rpc(self):
        """Setup RPC server for dfa agent."""

        endpoints = RpcCallBacks(self._vdpm, self._iptd)
        self.server = rpc.DfaRpcServer(self._qn, self._my_host, self._url,
                                       endpoints,
                                       exchange=constants.DFA_EXCHANGE)

    def start_rpc(self):
        self.server.start()
        LOG.debug('starting RPC server on the agent.')
        self.server.wait()

    def stop_rpc(self):
        self.server.stop()

    def start_rpc_task(self):
        thrd = utils.EventProcessingThread('Agent_RPC_Server',
                                           self, 'start_rpc')
        thrd.start()
        return thrd

    def start_iptables_task(self):
        thrd = self._iptd.create_thread()
        thrd.start()
        return thrd

    def start_tasks(self):
        rpc_thrd = self.start_rpc_task()
        self.agent_task_list.append(rpc_thrd)
        if (self._iptd):
            ipt_thrd = self.start_iptables_task()
            self.agent_task_list.append(ipt_thrd)


def save_my_pid(cfg):

    mypid = os.getpid()

    pid_path = cfg.dfa_log.pid_dir
    pid_file = cfg.dfa_log.pid_agent_file
    if pid_path and pid_file:
        try:
            if not os.path.exists(pid_path):
                os.makedirs(pid_path)
        except OSError:
            LOG.error('Fail to create %s', pid_path)
            return

        pid_file_path = os.path.join(pid_path, pid_file)

        LOG.debug('dfa_agent pid=%s', mypid)
        with open(pid_file_path, 'w') as fn:
            fn.write(str(mypid))


def main():

    # Setup logger
    cfg = config.CiscoDFAConfig().cfg
    logging.setup_logger('dfa_enabler', cfg)

    # Get pid of the process and save it.
    save_my_pid(cfg)

    # Create DFA agent object
    dfa_agent = DfaAgent(thishost, constants.DFA_AGENT_QUEUE)

    LOG.debug('Starting tasks in agent...')
    try:
        # Start all task in the agent.
        dfa_agent.start_tasks()

        # Endless loop
        while True:
            start = time.time()

            # Send heartbeat to controller, data includes:
            # - timestamp
            # - host name
            dfa_agent.send_heartbeat()

            # If the agent comes up for the fist time (could be after crash),
            # ask for the uplink info.
            if dfa_agent._need_uplink_info or (
               not dfa_agent.is_uplink_received()):
                dfa_agent.request_uplink_info()

            for trd in dfa_agent.agent_task_list:
                if not trd.am_i_active:
                    LOG.info("Thread %s is not active.", trd.name)

            end = time.time()
            delta = end - start
            eventlet.sleep(constants.HB_INTERVAL - delta)
    except Exception as e:
        dfa_agent.stop_rpc()
        LOG.exception('Exception %s is received', str(e))
        LOG.error('Exception %s is received', str(e))
        sys.exit("ERROR: %s" % str(e))

if __name__ == '__main__':
    sys.exit(main())



# Copyright 2015-2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six
import warnings

from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging

from networking_cisco._i18n import _

from networking_cisco.config import base

from networking_cisco.ml2_drivers.ucsm import constants as const

LOG = logging.getLogger(__name__)

""" Cisco UCS Manager ML2 Mechanism driver specific configuration.

Following are user configurable options for UCS Manager ML2 Mechanism
driver. The ucsm_username, ucsm_password, and ucsm_ip are
required options in single UCS Manager mode. A repetitive block starting
with ml2_cisco_ucsm_ip signals multi-UCSM configuration. When both are
present, the multi-UCSM config will only take effect.
"""

CONF = cfg.CONF


class EthPortType(types.String):

    def __call__(self, value):
        value = super(EthPortType, self).__call__(value)
        if not value.startswith(const.ETH_PREFIX):
            value = const.ETH_PREFIX + value
        return value


class VlanListType(types.List):

    def __call__(self, value):
        vlanlist = super(VlanListType, self).__call__(value)
        vlans = []
        for vlan in vlanlist:
            if '-' in vlan:
                start_vlan, sep, end_vlan = (vlan.partition('-'))
                vlans.extend(list(range(int(start_vlan.strip()),
                                        int(end_vlan.strip()) + 1, 1)))
            else:
                vlans.append(int(vlan))
        return vlans


class UCSTemplate(object):

    def __init__(self, path, name):
        self.path = path
        self.name = name

    def __eq__(self, other):
        return (isinstance(other, UCSTemplate) and self.path == other.path and
                self.name == other.name)


class SPTemplateListType(types.ConfigType):

    def __init__(self, type_name="SPTemplateList"):
        super(SPTemplateListType, self).__init__(type_name=type_name)

    def __call__(self, value):
        if isinstance(value, dict):
            return value

        templates = {}
        template_mappings = (value or "").split()

        for mapping in template_mappings:
            data = mapping.split(":")
            if len(data) != 3:
                raise cfg.Error(_('UCS Mech Driver: Invalid Service '
                                  'Profile Template config %s') % mapping)
            host_list = data[2].split(',')
            for host in host_list:
                templates[host] = UCSTemplate(data[0], data[1])
        return templates

    def _formatter(self, value):
        if isinstance(value, six.string_types):
            return value
        if isinstance(value, dict):
            template_to_hosts = {}
            for host, sptemplate in value.items():
                hosts = template_to_hosts.setdefault(
                    (sptemplate.path, sptemplate.name), [])
                hosts.append(host)
            template_configs = []
            for template, hosts in template_to_hosts.items():
                template_configs.append(
                    "%s:%s:%s" % (template[0], template[1], ','.join(hosts)))
            return ' '.join(template_configs)


class VNICTemplateListType(types.ConfigType):

    def __init__(self, type_name="VNICTemplateList"):
        super(VNICTemplateListType, self).__init__(type_name=type_name)

    def __call__(self, value):
        if isinstance(value, dict):
            return value

        templates = {}
        template_mappings = (value or "").split()

        for mapping in template_mappings:
            data = mapping.split(":")
            if len(data) != 3:
                raise cfg.Error(_("UCS Mech Driver: Invalid VNIC Template "
                                  "config: %s") % mapping)
            data[1] = data[1] or const.VNIC_TEMPLATE_PARENT_DN
            templates[data[0]] = UCSTemplate(data[1], data[2])
        return templates

    def _formatter(self, value):
        if isinstance(value, six.string_types):
            return value
        if isinstance(value, dict):
            templates = []
            for physnet, template in value.items():
                templates.append(
                    "%s:%s:%s" % (physnet, template.path, template.template))
            return ' '.join(templates)


ml2_cisco_ucsm_opts = [
    cfg.StrOpt('ucsm_ip',
               help=_('Cisco UCS Manager IP address. This is a required field '
                      'to communicate with a Cisco UCS Manager.')),
    cfg.ListOpt('supported_pci_devs',
                default=[const.PCI_INFO_CISCO_VIC_1240,
                         const.PCI_INFO_INTEL_82599],
                item_type=types.String(regex=r"^\w+:\w+$"),
                help=_('SR-IOV and VM-FEX vendors to be handled by the '
                       'driver. xxxx:yyyy represents vendor_id:product_id '
                       'of the PCI networking devices that the driver needs '
                       'to handle. It is implicit that the SR-IOV capable '
                       'devices specified here should be supported on the UCS '
                       'platform.')),
    cfg.BoolOpt('ucsm_https_verify',
                default=True,
                help=_('The UCSM driver will always perform SSL certificate '
                       'checking on the UCS Managers that it is connecting '
                       'to. This checking can be disabled by setting this '
                       'global configuration to False. '
                       'Disabling this check will leave the connection to UCS '
                       'Manager insecure and vulnerable to man-in-the-middle '
                       'attacks.')),
]

ml2_cisco_ucsm_common = [
    cfg.StrOpt('ucsm_username',
               help=_('Username for UCS Manager. This is a required field '
                      'to communicate with a Cisco UCS Manager.')),
    cfg.StrOpt('ucsm_password',
               secret=True,  # do not expose value in the logs
               help=_('Password for UCS Manager. This is a required field '
                      'to communicate with a Cisco UCS Manager.')),
    cfg.ListOpt('ucsm_virtio_eth_ports',
                default=[const.ETH_PREFIX + const.ETH0,
                         const.ETH_PREFIX + const.ETH1],
                item_type=EthPortType(),
                help=_('Ethernet port names to be used for virtio ports. '
                       'This config lets the Cloud Admin specify what ports '
                       'on the UCS Servers can be used for OpenStack virtual '
                       'port configuration. The names should match the '
                       'names on the UCS Manager.')),
    cfg.DictOpt('ucsm_host_list',
                help=_('Hostname to Service profile mapping for UCS Manager '
                       'controlled hosts. This Service profile '
                       'should not be associated with a Service Profile '
                       'Template. If the Service Profile is not specified '
                       'with a path, the driver assumes that it is at the '
                       'root level on the UCSM. For example: '
                       'Hostname1:Serviceprofile1, '
                       'Hostname2:Serviceprofile2')),
    cfg.StrOpt('sriov_qos_policy',
               default='${ml2_cisco_ucsm.sriov_qos_policy}',
               help=_('A pre-defined QoS policy name. This optional config '
                      'allows the cloud admin to pre-create a QoS policy on '
                      'the UCSM. If this config is present, the UCSM driver '
                      'will associate this QoS policy with every Port profile '
                      'it creates for SR-IOV ports.')),
    cfg.Opt('sp_template_list',
            type=SPTemplateListType(),
            default={},
            sample_default="<None>",
            help=_('Service Profile Template config for this UCSM. The '
                   'configuration to be provided should be a list where '
                   'each element in the list represents information for '
                   'a single Service Profile Template on that UCSM. Each '
                   'element is mapping of a Service Profile Template\'s '
                   'path, its name and a list of all UCS Servers '
                   'controlled by this template. For example:\n'
                   'sp_template_list = '
                   'SP_Template1_path:SP_Template1:Host1,Host2\n'
                   '                   '
                   'SP_Template2_path:SP_Template2:Host3,Host4\n'
                   'This is an optional config with no defaults')),
    cfg.Opt('vnic_template_list',
            type=VNICTemplateListType(),
            default={},
            sample_default="<None>",
            help=_('VNIC Profile Template config per UCSM. Allows the '
                   'cloud admin to specify a VNIC Template on the UCSM '
                   'that is attached to every vNIC connected to a specific '
                   'physical network. Each element in this list has 3 '
                   'parts: the physical network that is defined in neutron '
                   'configuration, the VNIC Template with its path in '
                   'UCSM, the vNIC on the UCS Servers that is connected to '
                   'this physical network. For example:\n'
                   'vnic_template_list = '
                   'physnet1:vnic_template_path1:vt1\n'
                   '                     '
                   'physnet2:vnic_template_path2:vt2\n'
                   'This is an optional config with no defaults.')),
]

sriov_opts = [
    base.RemainderOpt('network_name',
                      dest="network_vlans",
                      item_type=VlanListType(),
                      help=_('SR-IOV Multi-VLAN trunk config section is an '
                             'optional config section to accomodate the '
                             'scenario where an application using an SR-IOV '
                             'port to communicate would like to send traffic '
                             'on multiple application specific VLANs not '
                             'known to OpenStack. This config section is '
                             'applicable across all UCSMs specified as part '
                             'of the OpenStack cloud. The names of the '
                             'neutron networks on which the SR-IOV ports are '
                             'going to be created have to be known ahead of '
                             'time and should be associated with a list or '
                             'range of application VLANs using the following '
                             'format:\n'
                             '<neutron network name>=<comma separated list of '
                             'VLAN-ids or VLAN-id ranges> '
                             'For example:\ntest_network1=5,7-9')),
]

ucsms = base.SubsectionOpt(
    'ml2_cisco_ucsm_ip',
    dest='ucsms',
    help=_("Subgroups that allow you to specify the UCSMs to be "
           "managed by the UCSM ML2 driver."),
    subopts=ml2_cisco_ucsm_common)

CONF.register_opts(ml2_cisco_ucsm_opts, "ml2_cisco_ucsm")
CONF.register_opts(ml2_cisco_ucsm_common, "ml2_cisco_ucsm")
CONF.set_default("sriov_qos_policy", "", "ml2_cisco_ucsm")
CONF.register_opt(ucsms, "ml2_cisco_ucsm")
CONF.register_opts(sriov_opts, "sriov_multivlan_trunk")


def load_single_ucsm_config():
    # If no valid single configuration, skip this
    if not CONF.ml2_cisco_ucsm.ucsm_ip:
        return
    ucsm_ip = CONF.ml2_cisco_ucsm.ucsm_ip

    # Clear any previously loaded single ucsm config
    CONF.clear_override("ucsms", group="ml2_cisco_ucsm")

    if ucsm_ip in CONF.ml2_cisco_ucsm.ucsms:
        warnings.warn("UCSM %(ip)s is defined in the main [ml2_cisco_ucsm] "
                      "config group and the [ml2_cisco_ucsm_ip:%(ip)s] "
                      "config group. Using the configs from the"
                      "[ml2_cisco_ucsm_ip:%(ip)s] group. To remove this "
                      "warning remove the duplicated UCSM information from "
                      "the [ml2_cisco_ucsm] group in your config file."
                      % {"ip": ucsm_ip})
        return

    # Create a group to represent the single ucsms config
    CONF.register_opts(ml2_cisco_ucsm_common, "single_ucsm_config")

    # Inject config values from main ml2_cisco_ucsm group into the single ucsm
    # group
    for opt in ml2_cisco_ucsm_common:
        if opt.dest not in CONF.ml2_cisco_ucsm:
            continue
        CONF.set_override(opt.dest, CONF.ml2_cisco_ucsm[opt.dest],
                          group="single_ucsm_config")

    # Inject the single UCSM into the ucsms dictionary as an override so we can
    # clear it again later
    ucsms = dict(CONF.ml2_cisco_ucsm.ucsms)
    ucsms[ucsm_ip] = CONF.single_ucsm_config
    CONF.set_override("ucsms", ucsms, group="ml2_cisco_ucsm")


def add_sp_template_config_for_host(host, ucsm_ip,
                                    sp_template_path,
                                    sp_template):
    ucsms = CONF.ml2_cisco_ucsm.ucsms
    for ip, ucsm in ucsms.items():
        if ip == ucsm_ip:
            # NOTE(sambetts) Add this host to this UCSMs sp_template_list
            tp_list = dict(ucsm.sp_template_list)
            tp_list[host] = UCSTemplate(sp_template_path, sp_template)
            CONF.set_override("sp_template_list", tp_list, ucsm._group)
        else:
            # NOTE(sambetts) If this is not the UCSM we're adding the
            # service profile template for, then we need to check if this
            # UCSM has this host configured, if we do, remove it, because a
            # host can only be assigned to a single UCSM.
            if host in ucsm.sp_template_list:
                tp_list = dict(ucsm.sp_template_list)
                del tp_list[host]
                CONF.set_override("sp_template_list", tp_list, ucsm._group)


def update_sp_template_config(host_id, ucsm_ip,
                              sp_template_with_path):
    sp_template_info = sp_template_with_path.rsplit('/', 1)
    LOG.debug('SP Template path: %s SP Template: %s',
        sp_template_info[0], sp_template_info[1])
    add_sp_template_config_for_host(
        host_id, ucsm_ip, sp_template_info[0], sp_template_info[1])



# Copyright 2016 Cisco Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add support for UCSM Service Profile Templates.

Revision ID: 13bd9ebffbf5
Revises: 2e89171ea204
Create Date: 2016-01-27 13:12:12.805860

"""

# revision identifiers, used by Alembic.
revision = '13bd9ebffbf5'
down_revision = '2e89171ea204'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('ml2_ucsm_sp_templates',
        sa.Column('vlan_id', sa.Integer(), nullable=False),
        sa.Column('sp_template', sa.String(length=64), nullable=False),
        sa.Column('device_id', sa.String(length=64), nullable=False),
        sa.Column('updated_on_ucs', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('vlan_id', 'sp_template', 'device_id')
    )



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pprint
import sys

from oslo_config import cfg
import oslo_messaging
from oslo_utils import importutils

from neutron.common import config as common_config
from neutron.common import rpc as n_rpc
from neutron.common import topics

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.asr1k import (
    asr1k_cfg_syncer)
from networking_cisco.plugins.cisco.cfg_agent.device_drivers.asr1k import (
    asr1k_cfg_validator)
from networking_cisco.plugins.cisco.common import cisco_constants

manager = importutils.try_import('ncclient.manager')
# USAGE:
# python asr1k_auto_config_check.py --config-file /etc/neutron/neutron.conf


class CiscoDevMgrRPC(object):
    """Agent side of the device manager RPC API."""

    def __init__(self, topic, host):
        self.host = host
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def get_all_hosting_devices(self, context):
        """Get a list of all hosting devices."""
        cctxt = self.client.prepare()
        return cctxt.call(context,
                          'get_all_hosting_devices',
                          host=self.host)


class CiscoRoutingPluginRPC(object):
    """RoutingServiceHelper(Agent) side of the  routing RPC API."""

    def __init__(self, topic, host):
        self.host = host
        target = oslo_messaging.Target(topic=topic, version='1.1')
        self.client = n_rpc.get_client(target)

    def get_all_hosted_routers(self, context):
        """Make a remote process call to retrieve the sync data for
           routers that have been scheduled to a hosting device.

        :param context: session context
        """
        cctxt = self.client.prepare()
        return cctxt.call(context, 'cfg_sync_all_hosted_routers',
                          host=self.host)

    def get_hardware_router_type_id(self, context):
        """Get the ID for the ASR1k hardware router type."""
        cctxt = self.client.prepare()
        return cctxt.call(context,
                          'get_hardware_router_type_id',
                          host=self.host)


def get_nc_conn(hd):
    creds = hd['credentials']
    ncc_connection = manager.connect(host=hd['management_ip_address'],
                                     port=hd['protocol_port'],
                                     username=creds['user_name'],
                                     password=creds['password'],
                                     device_params={'name': "asr1k"},
                                     timeout=30)

    return ncc_connection


def main():

    conf = cfg.CONF

    common_config.init(sys.argv[1:])
    conf(project='neutron')

    host = conf.host
    devmgr_rpc = CiscoDevMgrRPC(cisco_constants.DEVICE_MANAGER_PLUGIN, host)
    plugin_rpc = CiscoRoutingPluginRPC(topics.L3PLUGIN, host)

    context = bc.ctxt.Context('', '')
    # TODO(create an admin context instead)

    hardware_router_type_id = plugin_rpc.get_hardware_router_type_id(context)
    print("Hardware router type ID: %s" % hardware_router_type_id)

    routers = plugin_rpc.get_all_hosted_routers(context)
    hosting_devs = devmgr_rpc.get_all_hosting_devices(context)

    print("ROUTERS: %s" % pprint.pformat(routers))

    for hd in hosting_devs['hosting_devices']:
        print("HOSTING DEVICE: %s, IP: %s\n-----------------" %
            (hd['id'], hd['management_ip_address']))

        if hd['template_id'] != hardware_router_type_id:
            continue

        conn = get_nc_conn(hd)

        cfg_cleaner = asr1k_cfg_syncer.ConfigSyncer(routers,
                                                    None,
                                                    hd,
                                                    test_mode=True)

        cfg_checker = asr1k_cfg_validator.ConfigValidator(routers,
                                                          hd,
                                                          conn)

        invalid_cfg = cfg_cleaner.delete_invalid_cfg(conn)
        missing_cfg = cfg_checker.process_routers_data(routers)

        print("Invalid Cfg: %s" % invalid_cfg)
        print("Missing Cfg: %s\n\n" % missing_cfg)

        conn.close_session()


if __name__ == "__main__":
    main()



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.common.htparser import HTParser
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routerrole

import re
import xml.etree.ElementTree as ET


ROUTER_ROLE_ATTR = routerrole.ROUTER_ROLE_ATTR

"""
Compares ASR running-config and neutron DB state, informs caller
if any configuration was missing from running-config.
"""


class ConfigValidator(object):
    def __init__(self, router_db_info, hosting_device_info, conn):
        self.hosting_device_info = hosting_device_info
        self.routers = router_db_info
        self.conn = conn

    def check_running_config(self):
        return self.process_routers_data(self.routers)

    def populate_segment_nat_dict(self, segment_nat_dict, routers):
        hd_id = self.hosting_device_info['id']
        for router in routers:
            if 'hosting_device' not in router:
                continue

            if router['hosting_device']['id'] != hd_id:
                continue

            # Mark which segments have NAT enabled
            # i.e., the segment is present on at least 1 router with
            # both external and internal networks present
            if 'gw_port' in router.keys():
                gw_port = router['gw_port']
                gw_segment_id = gw_port['hosting_info']['segmentation_id']
                if '_interfaces' in router.keys():
                    interfaces = router['_interfaces']
                    for intf in interfaces:
                        if intf['device_owner'] == \
                                bc.constants.DEVICE_OWNER_ROUTER_INTF:
                            intf_segment_id = \
                                intf['hosting_info']['segmentation_id']
                            segment_nat_dict[gw_segment_id] = True
                            segment_nat_dict[intf_segment_id] = True

    def process_routers_data(self, routers):
        hd_id = self.hosting_device_info['id']
        segment_nat_dict = {}
        #conn = self.driver._get_connection() #TODO(init ncclient properly)
        running_cfg = self.get_running_config(self.conn)
        parsed_cfg = HTParser(running_cfg)

        self.populate_segment_nat_dict(segment_nat_dict, routers)

        missing_cfg = []
        for router in routers:
            if 'hosting_device' not in router:
                continue

            if router['hosting_device']['id'] != hd_id:
                continue

            missing_cfg += self.check_router(router,
                                             parsed_cfg,
                                             segment_nat_dict)

        return missing_cfg

    def check_router(self, router, running_config, segment_nat_dict):
        if router[ROUTER_ROLE_ATTR] == cisco_constants.ROUTER_ROLE_GLOBAL:
            missing_cfg = self.check_global_router(router,
                                                   running_config,
                                                   segment_nat_dict)
        else:
            missing_cfg = self.check_tenant_router(router,
                                                   running_config,
                                                   segment_nat_dict)
        return missing_cfg

    def check_tenant_router(self, router, running_config, segment_nat_dict):
        missing_cfg = []
        missing_cfg += self.check_vrf(router, running_config)
        missing_cfg += self.check_nat_pool(router, running_config)
        missing_cfg += self.check_default_route(router, running_config)
        missing_cfg += self.check_acls(router, running_config)
        missing_cfg += self.check_fips(router, running_config)
        missing_cfg += self.check_interfaces(router,
                                             running_config,
                                             segment_nat_dict,
                                             is_external=False)
        return missing_cfg

    def check_global_router(self, router, running_config, segment_nat_dict):
        missing_cfg = []
        missing_cfg += self.check_interfaces(router,
                                             running_config,
                                             segment_nat_dict,
                                             is_external=True)
        return missing_cfg

    def get_running_config(self, conn):
        """Get the ASR1k's current running config.
        :return: Current IOS running config as multiline string
        """
        config = conn.get_config(source="running")
        if config:
            root = ET.fromstring(config._raw)
            running_config = root[0][0]
            rgx = re.compile("\r*\n+")
            ioscfg = rgx.split(running_config.text)
            return ioscfg

    def get_vrf_name(self, router):
        short_router_id = router['id'][0:6]
        return "nrouter-%s" % short_router_id

    def set_ip_cidr(self, intf):
        port_subnets = intf['subnets']
        subnet = port_subnets[0]
        prefixlen = netaddr.IPNetwork(subnet['cidr']).prefixlen
        intf['ip_cidr'] = "%s/%s" % (intf['fixed_ips'][0]['ip_address'],
                                     prefixlen)

    def get_interface_name_from_hosting_port(self, port):
        """
        generates the underlying subinterface name for a port
        e.g. Port-channel10.200
        """
        vlan = port['hosting_info']['segmentation_id']
        int_prefix = port['hosting_info']['physical_interface']
        return '%s.%s' % (int_prefix, vlan)

    def check_vrf(self, router, running_config):
        missing_cfg = []
        vrf_name = self.get_vrf_name(router)
        vrf_str = "vrf definition %s" % vrf_name
        vrf_substrs = [" address-family ipv4",
                       " address-family ipv6"]
        vrf_cfg = running_config.find_children(vrf_str)
        if not vrf_cfg:
            missing_cfg.append({"cfg": vrf_cfg})
        else:
            for substr in vrf_substrs:
                if substr not in vrf_cfg:
                    missing_cfg.append({"parent": vrf_str,
                                        "cfg": substr})
        return missing_cfg

    def check_nat_pool(self, router, running_config):
        missing_cfg = []

        if 'gw_port' not in router:
            return missing_cfg
        gw_port = router['gw_port']

        vrf_name = self.get_vrf_name(router)
        pool_name = "%s_nat_pool" % (vrf_name)
        pool_info = gw_port['nat_pool_info']
        pool_ip = pool_info['pool_ip']
        pool_net = netaddr.IPNetwork(pool_info['pool_cidr'])
        nat_pool_str = "ip nat pool %s %s %s netmask %s" % (pool_name,
                                                            pool_ip,
                                                            pool_ip,
                                                            pool_net.netmask)

        pool_cfg = running_config.find_lines(nat_pool_str)
        if not pool_cfg:
            missing_cfg.append({"cfg": nat_pool_str})

        if "_interfaces" in router:
            interfaces = router['_interfaces']
            for intf in interfaces:
                segment_id = intf['hosting_info']['segmentation_id']
                acl_name = "neutron_acl_%s" % segment_id

                nat_overload_str = "ip nat inside source list %s" \
                                   " pool %s vrf %s overload"
                nat_overload_str = nat_overload_str % (acl_name,
                                                       pool_name,
                                                       vrf_name)
                overload_cfg = running_config.find_lines(nat_overload_str)
                if not overload_cfg:
                    missing_cfg.append({"cfg": nat_overload_str})

        return missing_cfg

    def check_default_route(self, router, running_config):
        missing_cfg = []

        if 'gw_port' not in router:
            return missing_cfg

        vrf_name = self.get_vrf_name(router)

        gw_port = router['gw_port']
        ext_gw_ip = gw_port['subnets'][0]['gateway_ip']

        intf_name = self.get_interface_name_from_hosting_port(gw_port)

        route_str = "ip route vrf %s 0.0.0.0 0.0.0.0 %s %s" % (vrf_name,
                                                               intf_name,
                                                               ext_gw_ip)

        route_cfg = running_config.find_lines(route_str)
        if not route_cfg:
            missing_cfg.append({"cfg": route_str})

        return missing_cfg

    def check_acls(self, router, running_config):
        missing_cfg = []

        if "_interfaces" not in router:
            return missing_cfg

        interfaces = router["_interfaces"]
        for intf in interfaces:
            segment_id = intf['hosting_info']['segmentation_id']
            acl_name = "neutron_acl_%s" % segment_id
            self.set_ip_cidr(intf)
            internal_cidr = intf['ip_cidr']
            internal_net = netaddr.IPNetwork(internal_cidr).network
            net_mask = netaddr.IPNetwork(internal_cidr).hostmask

            acl_str = "ip access-list standard %s" % acl_name
            permit_str = " permit %s %s" % (internal_net, net_mask)

            acl_cfg = running_config.find_children(acl_str)
            if not acl_cfg:
                missing_cfg.append({"cfg": acl_str})
            else:
                if permit_str not in acl_cfg:
                    missing_cfg.append({"parent": acl_str, "cfg": permit_str})

        return missing_cfg

    def check_fips(self, router, running_config):
        missing_cfg = []

        if "_floatingips" not in router:
            return missing_cfg

        ex_gw_port = router['gw_port']
        vrf_name = self.get_vrf_name(router)

        fips = router["_floatingips"]
        for fip in fips:
            segment_id = ex_gw_port['hosting_info']['segmentation_id']
            hsrp_grp = ex_gw_port['nat_pool_info']['group']

            fip_str = "ip nat inside source static %s %s" \
                      " vrf %s redundancy neutron-hsrp-%s-%s"

            fip_str = fip_str % (fip['fixed_ip_address'],
                                 fip['floating_ip_address'],
                                 vrf_name,
                                 hsrp_grp,
                                 segment_id)

            fip_cfg = running_config.find_lines(fip_str)
            if not fip_cfg:
                missing_cfg.append({"cfg": fip_str})

        return missing_cfg

    def check_interfaces(self, router, running_config,
                         segment_nat_dict, is_external):
        missing_cfg = []

        if "_interfaces" not in router:
            return missing_cfg

        vrf_name = self.get_vrf_name(router)
        priority = router[ha.DETAILS][ha.PRIORITY]

        interfaces = router["_interfaces"]
        for intf in interfaces:
            segment_id = intf['hosting_info']['segmentation_id']
            intf_name = self.get_interface_name_from_hosting_port(intf)

            intf_str = "interface %s" % intf_name

            intf_cfg = running_config.find_children(intf_str)

            if not intf_cfg:
                missing_cfg.append({"cfg": intf_str})
            else:
                self.set_ip_cidr(intf)
                netmask = netaddr.IPNetwork(intf['ip_cidr']).netmask
                hsrp_vip = intf['fixed_ips'][0]['ip_address']
                port_ha_info = intf[ha.HA_INFO]
                hsrp_grp = port_ha_info['group']
                phys_ip = port_ha_info['ha_port']['fixed_ips'][0]['ip_address']

                sub_strs = [" description OPENSTACK_NEUTRON_INTF",
                            " encapsulation dot1Q %s" % segment_id,
                            " ip address %s %s" % (hsrp_vip, netmask),
                            " standby version 2",
                            " standby delay minimum 30 reload 60",
                            " standby %s priority %s" % (hsrp_grp, priority),
                            " standby %s ip %s" % (hsrp_grp, phys_ip),
                            " standby %s timers 1 3" % hsrp_grp]

                if not is_external:
                    sub_strs.append(" vrf forwarding %s" % vrf_name)

                if segment_id in segment_nat_dict:
                    if is_external:
                        sub_strs.append(" ip nat outside")
                    else:
                        sub_strs.append(" ip nat inside")

                for substr in sub_strs:
                    if substr not in intf_cfg:
                        missing_cfg.append({"parent": intf_str,
                                            "cfg": substr})

        return missing_cfg



# Copyright 2015 Cisco Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add missing tenant_id index

Revision ID: 11ba2d65c8de
Revises: 414ff6eb123a
Create Date: 2015-11-26 23:33:41.877280

"""

# revision identifiers, used by Alembic.
revision = '11ba2d65c8de'
down_revision = '414ff6eb123a'

from alembic import op


def upgrade():
    op.create_index(op.f('ix_cisco_router_ha_groups_tenant_id'),
                    'cisco_router_ha_groups', ['tenant_id'], unique=False)



# Copyright 2016 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from sqlalchemy import orm

from networking_cisco import backwards_compatibility as bc
from networking_cisco.ml2_drivers.ucsm import ucsm_model


class UcsmDbModel(object):
    def __init__(self):
        self.session = bc.get_writer_session()

    def is_port_profile_created(self, vlan_id, device_id):
        """Indicates if port profile has been created on UCS Manager."""
        entry = self.session.query(ucsm_model.PortProfile).filter_by(
            vlan_id=vlan_id, device_id=device_id).first()
        return entry and entry.created_on_ucs

    def get_port_profile_for_vlan(self, vlan_id, device_id):
        """Returns Vlan id associated with the port profile."""
        entry = self.session.query(ucsm_model.PortProfile).filter_by(
            vlan_id=vlan_id, device_id=device_id).first()
        return entry.profile_id if entry else None

    def add_port_profile(self, profile_name, vlan_id, device_id):
        """Adds a port profile and its vlan_id to the table."""
        if not self.get_port_profile_for_vlan(vlan_id, device_id):
            port_profile = ucsm_model.PortProfile(profile_id=profile_name,
                                                  vlan_id=vlan_id,
                                                  device_id=device_id,
                                                  created_on_ucs=False)
            with self.session.begin(subtransactions=True):
                self.session.add(port_profile)
            return port_profile

    def set_port_profile_created(self, vlan_id, profile_name, device_id):
        """Sets created_on_ucs flag to True."""
        with self.session.begin(subtransactions=True):
            port_profile = self.session.query(
                ucsm_model.PortProfile).filter_by(
                    vlan_id=vlan_id, profile_id=profile_name,
                    device_id=device_id).first()
            if port_profile:
                port_profile.created_on_ucs = True
                self.session.merge(port_profile)
            else:
                new_profile = ucsm_model.PortProfile(profile_id=profile_name,
                                          vlan_id=vlan_id,
                                          device_id=device_id,
                                          created_on_ucs=True)
                self.session.add(new_profile)

    def delete_vlan_entry(self, vlan_id):
        """Deletes entry for a vlan_id if it exists."""
        with self.session.begin(subtransactions=True):
            try:
                self.session.query(ucsm_model.PortProfile).filter_by(
                    vlan_id=vlan_id).delete()
            except orm.exc.NoResultFound:
                return

    def get_sp_template_vlan_entry(self, vlan_id, sp_template, ucsm_ip):
        return self.session.query(
            ucsm_model.ServiceProfileTemplate).filter_by(
                vlan_id=vlan_id,
                sp_template=sp_template,
                device_id=ucsm_ip).first()

    def add_service_profile_template(self, vlan_id, sp_template, ucsm_ip):
        """Adds an entry for a vlan_id on a SP template to the table."""
        if not self.get_sp_template_vlan_entry(vlan_id, sp_template, ucsm_ip):
            entry = ucsm_model.ServiceProfileTemplate(vlan_id=vlan_id,
                                                      sp_template=sp_template,
                                                      device_id=ucsm_ip,
                                                      updated_on_ucs=False)
            self.session.add(entry)

    def set_sp_template_updated(self, vlan_id, sp_template, device_id):
        """Sets update_on_ucs flag to True."""
        entry = self.get_sp_template_vlan_entry(vlan_id,
                                                sp_template,
                                                device_id)
        if entry:
            entry.updated_on_ucs = True
            self.session.merge(entry)
            return entry
        else:
            return False

    def delete_sp_template_for_vlan(self, vlan_id):
        """Deletes SP Template for a vlan_id if it exists."""
        with self.session.begin(subtransactions=True):
            try:
                self.session.query(
                    ucsm_model.ServiceProfileTemplate).filter_by(
                    vlan_id=vlan_id).delete()
            except orm.exc.NoResultFound:
                return

    def get_vnic_template_vlan_entry(self, vlan_id, vnic_template, ucsm_ip,
                                     physnet):
        return self.session.query(
            ucsm_model.VnicTemplate).filter_by(
                vlan_id=vlan_id,
                vnic_template=vnic_template,
                device_id=ucsm_ip,
                physnet=physnet).first()

    def add_vnic_template(self, vlan_id, ucsm_ip, vnic_template, physnet):
        """Adds an entry for a vlan_id on a SP template to the table."""
        if not self.get_vnic_template_vlan_entry(vlan_id, vnic_template,
            ucsm_ip, physnet):
            vnic_t = ucsm_model.VnicTemplate(vlan_id=vlan_id,
                                             vnic_template=vnic_template,
                                             device_id=ucsm_ip,
                                             physnet=physnet,
                                             updated_on_ucs=False)
            with self.session.begin(subtransactions=True):
                self.session.add(vnic_t)
            return vnic_t

    def set_vnic_template_updated(self, vlan_id, ucsm_ip, vnic_template,
                                  physnet):
        """Sets update_on_ucs flag to True for a Vnic Template entry."""
        with self.session.begin(subtransactions=True):
            entry = self.get_vnic_template_vlan_entry(vlan_id, vnic_template,
                        ucsm_ip, physnet)
            if entry:
                entry.updated_on_ucs = True
                self.session.merge(entry)
                return entry

    def delete_vnic_template_for_vlan(self, vlan_id):
        """Deletes VNIC Template for a vlan_id and physnet if it exists."""
        with self.session.begin(subtransactions=True):
            try:
                self.session.query(ucsm_model.VnicTemplate).filter_by(
                    vlan_id=vlan_id).delete()
            except orm.exc.NoResultFound:
                return

    def has_port_profile_to_delete(self, profile_name, device_id):
        """Returns True if port profile delete table containes PP."""
        count = self.session.query(ucsm_model.PortProfileDelete).filter_by(
            profile_id=profile_name, device_id=device_id).count()
        return count != 0

    def add_port_profile_to_delete_table(self, profile_name, device_id):
        """Adds a port profile to the delete table."""
        if not self.has_port_profile_to_delete(profile_name, device_id):
            port_profile = ucsm_model.PortProfileDelete(
                profile_id=profile_name, device_id=device_id)
            with self.session.begin(subtransactions=True):
                self.session.add(port_profile)
            return port_profile

    def get_all_port_profiles_to_delete(self):
        return self.session.query(ucsm_model.PortProfileDelete).all()

    def remove_port_profile_to_delete(self, profile_name, device_id):
        """Removes port profile to be deleted from table."""
        with self.session.begin(subtransactions=True):
            self.session.query(ucsm_model.PortProfileDelete).filter_by(
                profile_id=profile_name, device_id=device_id).delete()
